"""Serial Communication Interface Implementation
"""
import enum
import threading
import time
import logging
from collections import deque

import serial
import serial.tools.list_ports

from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.tm.definitions import TelemetryListT
from tmtccmd.utility.logger import get_console_logger
from dle_encoder import DleEncoder, STX_CHAR, ETX_CHAR, DleErrorCodes


LOGGER = get_console_logger()
SERIAL_FRAME_LENGTH = 256
DLE_FRAME_LENGTH = 1500
HEADER_BYTES_BEFORE_SIZE = 5


class SerialConfigIds(enum.Enum):
    from enum import auto

    SERIAL_PORT = auto()
    SERIAL_BAUD_RATE = auto()
    SERIAL_TIMEOUT = auto()
    SERIAL_COMM_TYPE = auto()
    SERIAL_FRAME_SIZE = auto()
    SERIAL_DLE_QUEUE_LEN = auto()
    SERIAL_DLE_MAX_FRAME_SIZE = auto()


class SerialCommunicationType(enum.Enum):
    """
    Right now, two serial communication methods are supported. One uses frames with a fixed size
    containing PUS packets and the other uses a simple ASCII based transport layer called DLE.
    If DLE is used, it is expected that the sender side encoded the packets with the DLE
    protocol. Any packets sent will also be encoded.
    """

    FIXED_FRAME_BASED = 1
    DLE_ENCODING = 2


# pylint: disable=arguments-differ
class SerialComIF(CommunicationInterface):
    """
    Communication Interface to use serial communication. This requires the PySerial library.
    """

    def __init__(
        self,
        com_if_key: str,
        tmtc_printer: TmTcPrinter,
        com_port: str,
        baud_rate: int,
        serial_timeout: float,
        ser_com_type: SerialCommunicationType = SerialCommunicationType.FIXED_FRAME_BASED,
    ):
        """
        Initiaze a serial communication handler.
        :param tmtc_printer: TMTC printer object. Can be used for diagnostic purposes, but main
        packet handling should be done by a separate thread.
        :param com_port: Specify COM port.
        :param baud_rate: Specify baud rate
        :param serial_timeout: Specify serial timeout
        :param ser_com_type: Specify how to handle serial reception
        """
        super().__init__(com_if_key=com_if_key, tmtc_printer=tmtc_printer)

        self.com_port = com_port
        self.baud_rate = baud_rate
        self.serial_timeout = serial_timeout
        self.serial = None
        self.encoder = None
        self.ser_com_type = ser_com_type
        if self.ser_com_type == SerialCommunicationType.FIXED_FRAME_BASED:
            # Set to default value.
            self.serial_frame_size = 256
        elif self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            self.encoder = DleEncoder()
            self.reception_thread = None
            self.reception_buffer = None
            self.dle_polling_active_event = None
            # Set to default value.
            self.dle_queue_len = 10
            self.dle_max_frame = 256
            self.dle_timeout = 0.01
            self.dle_encode_cr = True

    def __del__(self):
        if self.serial is not None:
            self.close()

    def set_fixed_frame_settings(self, serial_frame_size: int):
        self.serial_frame_size = serial_frame_size

    def set_dle_settings(
        self,
        dle_queue_len: int,
        dle_max_frame: int,
        dle_timeout: float,
        encode_cr: bool = True,
    ):
        self.dle_queue_len = dle_queue_len
        self.dle_max_frame = dle_max_frame
        self.dle_timeout = dle_timeout
        self.dle_encode_cr = encode_cr

    def initialize(self, args: any = None) -> any:
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            self.reception_buffer = deque(maxlen=self.dle_queue_len)
            self.dle_polling_active_event = threading.Event()

    def open(self, args: any = None) -> None:
        try:
            if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
                self.dle_polling_active_event.set()
                self.reception_thread = threading.Thread(
                    target=self.poll_dle_packets, daemon=True
                )
            self.serial = serial.Serial(
                port=self.com_port, baudrate=self.baud_rate, timeout=self.serial_timeout
            )
        except serial.SerialException:
            LOGGER.error("Serial Port opening failure!")
            raise IOError
        """
        Needs to be called by application code once for DLE mode!
        """
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            self.reception_thread.start()

    def close(self, args: any = None) -> None:
        try:
            if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
                self.dle_polling_active_event.clear()
                self.reception_thread.join(0.4)
            self.serial.close()
        except serial.SerialException:
            logging.warning("SERIAL Port could not be closed!")

    def send(self, data: bytearray):
        if self.ser_com_type == SerialCommunicationType.FIXED_FRAME_BASED:
            encoded_data = data
        elif self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            encoded_data = self.encoder.encode(source_packet=data, add_stx_etx=True)
        else:
            LOGGER.warning("This communication type was not implemented yet!")
            return
        self.serial.write(encoded_data)

    def receive(self, parameters: any = 0) -> TelemetryListT:
        packet_list = []
        if self.ser_com_type == SerialCommunicationType.FIXED_FRAME_BASED:
            if self.data_available():
                data = self.serial.read(self.serial_frame_size)
                pus_data_list = self.poll_pus_packets_fixed_frames(bytearray(data))
                for pus_packet in pus_data_list:
                    packet_list.append(pus_packet)
        elif self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            while self.reception_buffer:
                data = self.reception_buffer.pop()
                dle_retval, decoded_packet, read_len = self.encoder.decode(
                    source_packet=data
                )
                if dle_retval == DleErrorCodes.OK:
                    packet_list.append(decoded_packet)
                else:
                    LOGGER.warning("DLE decoder error!")
        else:
            LOGGER.warning("This communication type was not implemented yet!")
        return packet_list

    def data_available(self, timeout: float = 0, parameters: any = 0) -> int:
        elapsed_time = 0
        start_time = time.time()
        sleep_time = timeout / 3.0
        if self.ser_com_type == SerialCommunicationType.FIXED_FRAME_BASED:
            return self.data_available_fixed_frame(
                timeout=timeout, sleep_time=sleep_time
            )
        elif self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            if timeout > 0:
                while elapsed_time < timeout:
                    if self.reception_buffer:
                        return self.reception_buffer.__len__()
                    elapsed_time = time.time() - start_time
                    time.sleep(sleep_time)
            if self.reception_buffer:
                return self.reception_buffer.__len__()
        return 0

    def data_available_fixed_frame(self, timeout: float, sleep_time: float):
        if timeout > 0:
            start_time = time.time()
            elapsed_time = 0
            while elapsed_time < timeout:
                if self.serial.inWaiting() > 0:
                    return self.serial.inWaiting()
                elapsed_time = time.time() - start_time
                time.sleep(sleep_time)
        if self.serial.inWaiting() > 0:
            return self.serial.inWaiting()

    def poll_dle_packets(self):
        while True and self.dle_polling_active_event.is_set():
            # Poll permanently, but it is possible to join this thread every 200 ms
            self.serial.timeout = 0.2
            data = bytearray()
            byte = self.serial.read()
            if len(byte) == 1 and byte[0] == STX_CHAR:
                data.append(byte[0])
                self.serial.timeout = 0.1
                bytes_rcvd = self.serial.read_until(
                    serial.to_bytes([ETX_CHAR]), DLE_FRAME_LENGTH
                )
                if bytes_rcvd[len(bytes_rcvd) - 1] == ETX_CHAR:
                    data.extend(bytes_rcvd)
                    self.reception_buffer.appendleft(data)
            elif len(byte) >= 1:
                data.append(byte[0])
                data.extend(self.serial.read(self.serial.inWaiting()))
                # It is assumed that all packets are DLE encoded, so throw it away for now.
                LOGGER.info(f"Non DLE-Encoded data with length {len(data) + 1} found..")

    @staticmethod
    def poll_pus_packets_fixed_frames(data: bytearray) -> list:
        pus_data_list = []
        if len(data) == 0:
            return pus_data_list

        payload_length = data[4] << 8 | data[5]
        packet_size = payload_length + 7
        if payload_length == 0:
            return []
        read_size = len(data)
        pus_data = data[0:packet_size]
        pus_data_list.append(pus_data)

        SerialComIF.read_multiple_packets(data, packet_size, read_size, pus_data_list)
        return pus_data_list

    @staticmethod
    def read_multiple_packets(
        data: bytearray, start_index: int, frame_size: int, pus_data_list: list
    ):
        while start_index < frame_size:
            start_index = SerialComIF.parse_next_packets(
                data, start_index, frame_size, pus_data_list
            )

    @staticmethod
    def parse_next_packets(
        data: bytearray, start_index: int, frame_size: int, pus_data_list: list
    ) -> int:
        next_payload_len = data[start_index + 4] << 8 | data[start_index + 5]
        if next_payload_len == 0:
            end_index = frame_size
            return end_index
        next_packet_size = next_payload_len + 7
        # remaining_size = frame_size - start_index

        if next_packet_size > SERIAL_FRAME_LENGTH:
            LOGGER.error(
                "PUS Polling: Very large packet detected, packet splitting not implemented yet!"
            )
            LOGGER.error("Detected Size: " + str(next_packet_size))
            end_index = frame_size
            return end_index

        end_index = start_index + next_packet_size
        pus_data = data[start_index:end_index]
        pus_data_list.append(pus_data)
        return end_index



from threading import Lock

__GLOBALS_DICT = dict()
__LOCK_TIMEOUT = 50
__GLOBALS_LOCK = Lock()


def get_global(global_param_id: int, lock: bool = False):
    global __LOCK_TIMEOUT, __GLOBALS_DICT
    if lock:
        __GLOBALS_LOCK.acquire(timeout=__LOCK_TIMEOUT)
    global_param = __GLOBALS_DICT.get(global_param_id)
    if lock:
        __GLOBALS_LOCK.release()
    return global_param


def update_global(global_param_id: int, parameter: any, lock: bool = False):
    global __LOCK_TIMEOUT, __GLOBALS_DICT
    if lock:
        __GLOBALS_LOCK.acquire(timeout=__LOCK_TIMEOUT)
    __GLOBALS_DICT[global_param_id] = parameter
    if lock:
        __GLOBALS_LOCK.release()


def lock_global_pool() -> bool:
    global __LOCK_TIMEOUT, __GLOBALS_LOCK
    """Lock the global objects. This is important if the values are changed. Don't forget to unlock the pool
    after finishing work with the globals!
    :param timeout_seconds: Attempt to lock for this many second. Default value -1 blocks permanently until lock is
    released.
    :return: Returns whether lock was locked or not.
    """
    return __GLOBALS_LOCK.acquire(__LOCK_TIMEOUT)


def unlock_global_pool():
    global __GLOBALS_LOCK
    """Releases the lock so other objects can use the global pool as well"""
    return __GLOBALS_LOCK.release()


def set_lock_timeout(timeout: float):
    global __LOCK_TIMEOUT
    """Set the timeout for the globals manager lock which can ensure thread-safety"""
    __LOCK_TIMEOUT = timeout



#!/usr/bin/python3.8
"""
@file   tmtcc_sequential_sender_receiver.py
@date   01.11.2019
@brief  Used to send multiple TCs in sequence and listen for replies after each sent TC
"""
import sys
import time

from tmtccmd.sendreceive.cmd_sender_receiver import CommandSenderReceiver
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.sendreceive.tm_listener import TmListener
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.tc.definitions import TcQueueT

LOGGER = get_console_logger()


class SequentialCommandSenderReceiver(CommandSenderReceiver):
    """Specific implementation of CommandSenderReceiver to send multiple telecommands in sequence"""

    def __init__(
        self,
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tm_handler: CcsdsTmHandler,
        apid: int,
        tm_listener: TmListener,
        tc_queue: TcQueueT,
    ):
        """
        :param com_if:          CommunicationInterface object, passed on to CommandSenderReceiver
        :param tm_listener:     TmListener object which runs in the background and receives
                                all Telemetry
        :param tmtc_printer:    TmTcPrinter object, passed on to CommandSenderReceiver for
                                this time period
        """
        super().__init__(
            com_if=com_if,
            tmtc_printer=tmtc_printer,
            tm_listener=tm_listener,
            tm_handler=tm_handler,
            apid=apid,
        )
        self._tc_queue = tc_queue
        self.__all_replies_received = False

    def send_queue_tc_and_receive_tm_sequentially(self):
        """Primary function which is called for sequential transfer.
        :return:
        """
        self._tm_listener.set_listener_mode(TmListener.ListenerModes.SEQUENCE)
        # tiny delay for pus_tm listener
        time.sleep(0.05)
        if self._tc_queue:
            try:
                # Set to true for first packet, otherwise nothing will be sent.
                self._reply_received = True
                self.__handle_tc_sending()
            except (KeyboardInterrupt, SystemExit):
                LOGGER.info("Keyboard Interrupt.")
                sys.exit()
        else:
            LOGGER.warning("Supplied TC queue is empty!")

    def __handle_tc_sending(self):
        while not self.__all_replies_received:
            while not self._tc_queue.__len__() == 0:
                self.__check_for_reply()
                self.__check_next_tc_send()
                if self._tc_queue.__len__() == 0:
                    self._start_time = time.time()
                    break
                time.sleep(0.2)
            if not self._reply_received:
                self.__check_for_reply()
                self._check_for_timeout()
            if self._reply_received:
                self.__all_replies_received = True
                break
            time.sleep(0.2)
        self._tm_listener.set_mode_op_finished()
        LOGGER.info("SequentialSenderReceiver: All replies received!")

    def __check_for_reply(self):
        if self._tm_listener.reply_event():
            self._reply_received = True
            self._tm_listener.clear_reply_event()
            packet_queue = self._tm_listener.retrieve_ccsds_tm_packet_queue(
                apid=self._apid, clear=True
            )
            self._tm_handler.handle_ccsds_packet_queue(
                apid=self._apid, packet_queue=packet_queue
            )
        # This makes reply reception more responsive
        elif self._tm_listener.tm_packets_available():
            packet_queue = self._tm_listener.retrieve_ccsds_tm_packet_queue(
                apid=self._apid, clear=True
            )
            self._tm_handler.handle_ccsds_packet_queue(
                apid=self._apid, packet_queue=packet_queue
            )

    def __check_next_tc_send(self):
        if self.wait_period_ongoing():
            return
        # this flag is set in the separate receiver thread too
        if self._reply_received:
            if self.__send_next_telecommand():
                self._reply_received = False
        # just calculate elapsed time if start time has already been set (= command has been sent)
        else:
            self._check_for_timeout()

    def __send_next_telecommand(self) -> bool:
        """Sends the next telecommand and returns whether an actual telecommand was sent"""
        tc_queue_tuple = self._tc_queue.pop()
        if self.check_queue_entry(tc_queue_tuple):
            self._start_time = time.time()
            pus_packet, pus_packet_info = tc_queue_tuple
            self._com_if.send(pus_packet)
            return True
        # queue empty.
        elif not self._tc_queue:
            # Special case: Last queue entry is not a Telecommand
            self._reply_received = True
            # Another specal case: Last queue entry is to wait.
            if self._wait_period > 0:
                self.wait_period_ongoing(True)
                self.__all_replies_received = True
            return False
        else:
            # If the queue entry was not a telecommand, send next telecommand
            self.__check_next_tc_send()
            return True



"""
Argument parser modules for the TMTC commander core
"""
import argparse
import sys

from tmtccmd.config.definitions import CoreModeList, ServiceOpCodeDictT
from tmtccmd.utility.logger import get_console_logger


LOGGER = get_console_logger()


def parse_input_arguments(
    print_known_args: bool = False, print_unknown_args: bool = False
) -> argparse.Namespace:
    try:
        from tmtccmd.config.hook import get_global_hook_obj

        hook_obj = get_global_hook_obj()
        args = hook_obj.custom_args_parsing()
        if args is None:
            return parse_default_input_arguments(print_known_args, print_unknown_args)
    except ImportError:
        return parse_default_input_arguments(print_known_args, print_unknown_args)


def parse_default_input_arguments(
    print_known_args: bool = False, print_unknown_args: bool = False
):
    """Parses all input arguments
    :return: Input arguments contained in a special namespace and accessable by args.<variable>
    """
    from tmtccmd.utility.conf_util import AnsiColors

    descrip_text = (
        f"{AnsiColors.GREEN}TMTC Client Command Line Interface\n"
        f"{AnsiColors.RESET}This application provides generic components to execute "
        f"TMTC commanding.\n"
        f"The developer is expected to specify the packaged telecommands for a given\n"
        "service and operation code combination. The developer is also expected\n"
        "to implement the handling of telemetry. All these tasks can be done by implementing\n"
        "a hook object and passing it to the core."
    )
    arg_parser = argparse.ArgumentParser(
        description=descrip_text, formatter_class=argparse.RawTextHelpFormatter
    )

    add_default_mode_arguments(arg_parser)
    add_default_com_if_arguments(arg_parser)
    add_generic_arguments(arg_parser)

    # TODO: Only add this if TMTC commander is configured for Ethernet?
    add_ethernet_arguments(arg_parser)

    arg_parser.add_argument(
        "--tc_timeout_factor",
        type=float,
        help="TC Timeout Factor. Multiplied with "
        "TM Timeout, TC sent again after this time period. Default: 3.5",
        default=3.5,
    )
    arg_parser.add_argument(
        "-r",
        "--raw_data_print",
        help="Supply -r to print all raw TM data directly",
        action="store_true",
    )
    arg_parser.add_argument(
        "-d",
        "--short_display_mode",
        help="Supply -d to print short output",
        action="store_true",
    )
    arg_parser.add_argument(
        "--hk",
        dest="print_hk",
        help="Supply -k or --hk to print HK data",
        action="store_true",
    )
    arg_parser.add_argument(
        "--rs",
        dest="resend_tc",
        help="Specify whether TCs are sent again after timeout",
        action="store_true",
    )

    if len(sys.argv) == 1:
        LOGGER.info(
            "No input arguments specified. Run with -h to get list of arguments"
        )

    args, unknown = arg_parser.parse_known_args()

    if print_known_args:
        LOGGER.info("Printing known arguments:")
        for argument in vars(args):
            LOGGER.debug(argument + ": " + str(getattr(args, argument)))
    if print_unknown_args:
        LOGGER.info("Printing unknown arguments:")
        for argument in unknown:
            LOGGER.info(argument)

    args_post_processing(args, unknown)
    return args


def add_generic_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument(
        "-s", "--service", type=str, help="Service to test", default=None
    )
    arg_parser.add_argument(
        "-o",
        "--op_code",
        help="Operation code, which is passed to the TC packer functions",
        default=None,
    )
    arg_parser.add_argument(
        "-l",
        "--listener",
        help="Determine whether the listener mode will be active after performing the operation",
        action="store_false",
    )
    arg_parser.add_argument(
        "-t",
        "--tm_timeout",
        type=float,
        help="TM Timeout when listening to verification sequence."
        " Default: 5 seconds",
        default=5.0,
    )
    arg_parser.add_argument(
        "--nl",
        dest="print_log",
        help="Supply --nl to suppress print output to log files.",
        action="store_false",
    )
    arg_parser.add_argument(
        "--np",
        dest="print_tm",
        help="Supply --np to suppress print output to console.",
        action="store_false",
    )


def add_default_mode_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config.definitions import CoreModeList, CoreModeStrings

    help_text = "Core Modes. Default: seqcmd\n"
    seq_help = (
        f"{CoreModeList.SEQUENTIAL_CMD_MODE} or "
        f"{CoreModeStrings[CoreModeList.SEQUENTIAL_CMD_MODE]}: "
        f"Sequential Command Mode\n"
    )
    listener_help = (
        f"{CoreModeList.LISTENER_MODE} or {CoreModeStrings[CoreModeList.LISTENER_MODE]}: "
        f"Listener Mode\n"
    )
    gui_help = (
        f"{CoreModeList.GUI_MODE} or "
        f"{CoreModeStrings[CoreModeList.GUI_MODE]}: "
        f"GUI mode\n"
    )
    help_text += seq_help + listener_help + gui_help
    arg_parser.add_argument(
        "-m",
        "--mode",
        type=str,
        help=help_text,
        default=CoreModeStrings[CoreModeList.SEQUENTIAL_CMD_MODE],
    )


def add_default_com_if_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config.definitions import CoreComInterfacesDict, CoreComInterfaces

    help_text = (
        "Core Communication Interface. If this is not specified, the commander core\n"
        "will try to extract it from the JSON or prompt it from the user.\n"
    )
    dummy_line = (
        f"{CoreComInterfacesDict[CoreComInterfaces.DUMMY.value]}: Dummy Interface\n"
    )
    udp_line = (
        f"{CoreComInterfacesDict[CoreComInterfaces.TCPIP_UDP.value]}: " f"UDP client\n"
    )
    ser_dle_line = (
        f"{CoreComInterfacesDict[CoreComInterfaces.SERIAL_DLE.value]}: "
        f"Serial with DLE transport layer\n"
    )
    ser_fixed_line = (
        f"{CoreComInterfacesDict[CoreComInterfaces.SERIAL_FIXED_FRAME.value]}: "
        f"Serial with fixed frames\n"
    )
    ser_qemu_line = (
        f"{CoreComInterfacesDict[CoreComInterfaces.SERIAL_QEMU.value]}: "
        f"QEMU serial interface\n"
    )
    help_text += dummy_line + ser_dle_line + udp_line + ser_fixed_line + ser_qemu_line
    arg_parser.add_argument(
        "-c",
        "--com_if",
        type=str,
        help=help_text,
        default=CoreComInterfaces.UNSPECIFIED.value,
    )


def add_ethernet_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument(
        "--clientIP", help="Client(Computer) IP. Default:''", default=""
    )
    arg_parser.add_argument(
        "--boardIP", help="Board IP. Default: Localhost 127.0.0.1", default="127.0.0.1"
    )


def args_post_processing(args, unknown: list) -> None:
    """Handles the parsed arguments.
    :param args: Namespace objects (see https://docs.python.org/dev/library/argparse.html#argparse.Namespace)
    :param unknown: List of unknown parameters.
    :return: None
    """
    if len(unknown) > 0:
        print("Unknown arguments detected: " + str(unknown))
    if len(sys.argv) > 1:
        handle_unspecified_args(args)
    if len(sys.argv) == 1:
        handle_empty_args(args)


def handle_unspecified_args(args) -> None:
    """If some arguments are unspecified, they are set here with (variable) default values.
    :param args:
    :return: None
    """
    from tmtccmd.config.hook import get_global_hook_obj
    from tmtccmd.config.definitions import CoreModeStrings

    if args.tm_timeout is None:
        args.tm_timeout = 5.0
    if args.mode is None:
        args.mode = CoreModeStrings[CoreModeList.SEQUENTIAL_CMD_MODE]
    service_op_code_dict = dict()
    if args.service is None or args.op_code is None:
        hook_obj = get_global_hook_obj()
        service_op_code_dict = hook_obj.get_service_op_code_dictionary()
    if args.service is None:
        if args.mode == CoreModeStrings[CoreModeList.SEQUENTIAL_CMD_MODE]:
            LOGGER.info("No service argument (-s) specified, prompting from user..")
            # Try to get the service list from the hook base and prompt service from user
            args.service = prompt_service(service_op_code_dict)
            if args.op_code is None:
                current_service = args.service
                args.op_code = prompt_op_code(
                    service_op_code_dict=service_op_code_dict, service=current_service
                )
    elif args.op_code is None:
        current_service = args.service
        args.op_code = prompt_op_code(
            service_op_code_dict=service_op_code_dict, service=current_service
        )


def handle_empty_args(args) -> None:
    """If no args were supplied, request input from user directly.
    :param args:
    :return:
    """
    LOGGER.info("No arguments specified..")
    handle_unspecified_args(args=args)


def prompt_service(service_op_code_dict: ServiceOpCodeDictT) -> str:
    service_adjustment = 20
    info_adjustment = 30
    horiz_line_num = service_adjustment + info_adjustment + 3
    horiz_line = horiz_line_num * "-"
    service_string = "Service".ljust(service_adjustment)
    info_string = "Information".ljust(info_adjustment)
    while True:
        LOGGER.info(f"{service_string} | {info_string}")
        LOGGER.info(horiz_line)
        for service_entry in service_op_code_dict.items():
            adjusted_service_entry = service_entry[0].ljust(service_adjustment)
            adjusted_service_info = service_entry[1][0].ljust(info_adjustment)
            LOGGER.info(f"{adjusted_service_entry} | {adjusted_service_info}")
        service_string = input("Please select a service by specifying the key: ")
        if service_string in service_op_code_dict:
            LOGGER.info(f"Selected service: {service_string}")
            return service_string
        else:
            LOGGER.warning("Invalid key, try again")


def prompt_op_code(service_op_code_dict: ServiceOpCodeDictT, service: str) -> str:
    op_code_adjustment = 16
    info_adjustment = 34
    horz_line_num = op_code_adjustment + info_adjustment + 3
    horiz_line = horz_line_num * "-"
    op_code_info_str = "Operation Code".ljust(op_code_adjustment)
    info_string = "Information".ljust(info_adjustment)
    while True:
        LOGGER.info(f"{op_code_info_str} | {info_string}")
        LOGGER.info(horiz_line)
        if service in service_op_code_dict:
            op_code_dict = service_op_code_dict[service][1]
            for op_code_entry in op_code_dict.items():
                adjusted_op_code_entry = op_code_entry[0].ljust(op_code_adjustment)
                adjusted_op_code_info = op_code_entry[1][0].ljust(info_adjustment)
                LOGGER.info(f"{adjusted_op_code_entry} | {adjusted_op_code_info}")
            op_code_string = input(
                "Please select an operation code by specifying the key: "
            )
            if op_code_string in op_code_dict:
                LOGGER.info(f"Selected op code: {op_code_string}")
                return op_code_string
            else:
                LOGGER.warning("Invalid key, try again")
        else:
            LOGGER.warning(
                "Service not in dictionary. Setting default operation code 0"
            )
            return "0"



"""
@brief      This module is used to set up the global loggers
"""
import logging
import os
import sys


TMTC_LOGGER_NAME = "TMTC Console Logger"
TMTC_FILE_LOGGER_NAME = "TMTC File Logger"
ERROR_LOG_FILE_NAME = "tmtc_error.log"
__CONSOLE_LOGGER_SET_UP = False
__FILE_LOGER_SET_UP = False


# pylint: disable=arguments-differ
# pylint: disable=too-few-public-methods
class InfoFilter(logging.Filter):
    """Filter object, which is used so that only INFO and DEBUG messages are printed to stdout."""

    def filter(self, rec):
        if rec.levelno == logging.INFO:
            return rec.levelno
        return None


class DebugFilter(logging.Filter):
    """Filter object, which is used so that only DEBUG messages are printed to stdout."""

    def filter(self, rec):
        if rec.levelno == logging.DEBUG:
            return rec.levelno
        return None


def set_tmtc_console_logger() -> logging.Logger:
    """Sets the LOGGER object which will be used globally. This needs to be called before
    using the logger.
    :return:    Returns the instance of the global logger
    """
    global __CONSOLE_LOGGER_SET_UP
    logger = logging.getLogger(TMTC_LOGGER_NAME)
    logger.setLevel(level=logging.DEBUG)

    # Use colorlog for now because it allows more flexibility and custom messages
    # for different levels
    set_up_colorlog_logger(logger=logger)

    # set_up_coloredlogs_logger(logger=logger)

    file_format = logging.Formatter(
        fmt="%(levelname)-8s: %(asctime)s.%(msecs)03d | [%(filename)s:%(lineno)d] | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    console_error_handler = logging.StreamHandler(stream=sys.stderr)
    console_error_handler.setLevel(logging.WARNING)

    try:
        error_file_handler = logging.FileHandler(
            filename=f"log/{ERROR_LOG_FILE_NAME}", encoding="utf-8", mode="w"
        )
    except FileNotFoundError:
        os.mkdir("log")
        error_file_handler = logging.FileHandler(
            filename=f"log/{ERROR_LOG_FILE_NAME}", encoding="utf-8", mode="w"
        )
    error_file_handler.setLevel(level=logging.WARNING)
    error_file_handler.setFormatter(file_format)
    logger.addHandler(error_file_handler)

    __CONSOLE_LOGGER_SET_UP = True
    return logger


def set_up_coloredlogs_logger(logger: logging.Logger):
    try:
        import coloredlogs

        coloredlogs.install(
            level="INFO",
            logger=logger,
            milliseconds=True,
            fmt="%(asctime)s.%(msecs)03d %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    except ImportError:
        print("Please install coloredlogs package first")


def set_up_colorlog_logger(logger: logging.Logger):
    from colorlog import ColoredFormatter, StreamHandler

    generic_format = ColoredFormatter(
        fmt="%(log_color)s%(levelname)-8s %(cyan)s%(asctime)s.%(msecs)03d %(reset)s%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    fault_format = ColoredFormatter(
        fmt="%(log_color)s%(levelname)-8s %(cyan)s%(asctime)s.%(msecs)03d "
        "[%(filename)s:%(lineno)d] %(reset)s%(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    file_format = logging.Formatter(
        fmt="%(levelname)-8s: %(asctime)s.%(msecs)03d [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    console_info_handler = StreamHandler(stream=sys.stdout)
    console_info_handler.setLevel(logging.INFO)
    console_info_handler.addFilter(InfoFilter())

    console_debug_handler = logging.StreamHandler(stream=sys.stdout)
    console_debug_handler.setLevel(logging.DEBUG)
    console_debug_handler.addFilter(DebugFilter())

    console_error_handler = logging.StreamHandler(stream=sys.stderr)
    console_error_handler.setLevel(logging.WARNING)

    try:
        error_file_handler = logging.FileHandler(
            filename=f"log/{ERROR_LOG_FILE_NAME}", encoding="utf-8", mode="w"
        )
    except FileNotFoundError:
        os.mkdir("log")
        error_file_handler = logging.FileHandler(
            filename=f"log/{ERROR_LOG_FILE_NAME}", encoding="utf-8", mode="w"
        )

    error_file_handler.setLevel(level=logging.WARNING)
    error_file_handler.setFormatter(file_format)
    console_info_handler.setFormatter(generic_format)
    console_debug_handler.setFormatter(fault_format)
    console_error_handler.setFormatter(fault_format)
    logger.addHandler(error_file_handler)
    logger.addHandler(console_info_handler)
    logger.addHandler(console_debug_handler)
    logger.addHandler(console_error_handler)


def get_console_logger(set_up_logger: bool = False) -> logging.Logger:
    global __CONSOLE_LOGGER_SET_UP
    """Get the global console logger instance. Error logs will still be saved to an error file
    """
    logger = logging.getLogger(TMTC_LOGGER_NAME)
    if set_up_logger and not __CONSOLE_LOGGER_SET_UP:
        __CONSOLE_LOGGER_SET_UP = True
        set_tmtc_console_logger()
    return logger



from unittest import TestCase
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.runner import (
    run_tmtc_commander,
    initialize_tmtc_commander,
    get_default_tmtc_backend,
)

from tests.backend_mock import create_backend_mock, create_frontend_mock
from tests.hook_obj_mock import create_hook_mock


class TestTmtcRunner(TestCase):
    def test_tmtc_runner(self):
        hook_base = create_hook_mock()
        tm_handler = CcsdsTmHandler(tmtc_printer=None)
        backend_mock = create_backend_mock(tm_handler=tm_handler)
        initialize_tmtc_commander(hook_object=hook_base)
        run_tmtc_commander(False, False, True, tmtc_backend=backend_mock)
        backend_mock.start_listener.assert_called_with()
        backend_mock.initialize.assert_called_with()

        frontend_mock = create_frontend_mock()
        run_tmtc_commander(
            True, False, True, tmtc_backend=backend_mock, tmtc_frontend=frontend_mock
        )
        frontend_mock.start.assert_called_once()
        qt_app = frontend_mock.start.call_args[0][0]
        self.assertTrue(qt_app is None)
        default_backend = get_default_tmtc_backend(
            hook_obj=hook_base, tm_handler=tm_handler, json_cfg_path="tmtc_config.json"
        )
        self.assertTrue(default_backend is not None)

    def test_errors(self):
        self.assertRaises(ValueError, initialize_tmtc_commander, None)
        self.assertRaises(TypeError, run_tmtc_commander)
        self.assertRaises(RuntimeError, run_tmtc_commander, False)



# -*- coding: utf-8 -*-
"""Deserialize PUS Service 1 Verification TM
"""
from __future__ import annotations
from abc import abstractmethod
import struct
from typing import Deque

from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.tm import PusVersion, PusTelemetry
from spacepackets.ecss.service_1_verification import Service1TM

from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class Service1TMExtended(PusTmBase, PusTmInfoBase, Service1TM):
    """Service 1 TM class representation. Can be used to deserialize raw service 1 packets."""

    def __init__(
        self,
        subservice: int,
        time: CdsShortTimestamp = None,
        tc_packet_id: int = 0,
        tc_psc: int = 0,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        Service1TM.__init__(
            self,
            subservice=subservice,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=self.pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=self.pus_tm)

    @classmethod
    def __empty(cls) -> Service1TMExtended:
        return cls(subservice=0)

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service1TMExtended:
        """Parse a service 1 telemetry packet

        :param raw_telemetry:
        :param pus_version:
        :raises ValueError: Raw telemetry too short
        :return:
        """
        service_1_tm = cls.__empty()
        service_1_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        tm_data = service_1_tm.tm_data
        if len(tm_data) < 4:
            LOGGER.warning("TM data less than 4 bytes!")
            raise ValueError
        service_1_tm.tc_packet_id = tm_data[0] << 8 | tm_data[1]
        service_1_tm.tc_psc = tm_data[2] << 8 | tm_data[3]
        service_1_tm.tc_ssc = service_1_tm.tc_psc & 0x3FFF
        if service_1_tm.subservice % 2 == 0:
            service_1_tm._handle_failure_verification()
        else:
            service_1_tm._handle_success_verification()
        return service_1_tm

    @abstractmethod
    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(str(hex(self.tc_packet_id)))
        content_list.append(str(self.tc_ssc))
        if self.has_tc_error_code:
            if self.is_step_reply:
                content_list.append(str(self.step_number))
            content_list.append(str(hex(self.error_code)))
            content_list.append(
                f"hex {self.error_param_1:04x} dec {self.error_param_1}"
            )
            content_list.append(
                f"hex {self.error_param_2:04x} dec {self.error_param_2}"
            )
        elif self.is_step_reply:
            content_list.append(str(self.step_number))

    @abstractmethod
    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("TC Packet ID")
        header_list.append("TC SSC")
        if self.has_tc_error_code:
            if self.is_step_reply:
                header_list.append("Step Number")
            header_list.append("Return Value")
            header_list.append("Error Param 1")
            header_list.append("Error Param 2")
        elif self.is_step_reply:
            header_list.append("Step Number")

    def _handle_failure_verification(self):
        """Handle parsing a verification failure packet, subservice ID 2, 4, 6 or 8"""
        super()._handle_failure_verification()
        self.set_packet_info("Failure Verficiation")
        subservice = self.pus_tm.subservice
        if subservice == 2:
            self.append_packet_info(" : Acceptance failure")
        elif subservice == 4:
            self.append_packet_info(" : Start failure")
        elif subservice == 6:
            self.append_packet_info(" : Step Failure")
        elif subservice == 8:
            self.append_packet_info(" : Completion Failure")

    def _handle_success_verification(self):
        super()._handle_success_verification()
        self.set_packet_info("Success Verification")
        if self.subservice == 1:
            self.append_packet_info(" : Acceptance success")
        elif self.subservice == 3:
            self.append_packet_info(" : Start success")
        elif self.subservice == 5:
            self.append_packet_info(" : Step Success")
        elif self.subservice == 7:
            self.append_packet_info(" : Completion success")


PusVerifQueue = Deque[Service1TM]



"""Base class for implementation of PUS Service 2 handling.
"""
from __future__ import annotations
from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry

from tmtccmd.tm.base import PusTmInfoBase, PusTmBase


class Service2TM(PusTmInfoBase, PusTmBase):
    def __init__(
        self,
        subservice: int,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        pus_tm = PusTelemetry(
            service=2,
            subservice=subservice,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        PusTmBase.__init__(self, pus_tm=pus_tm)
        self.set_packet_info("Raw Commanding Reply")

    @classmethod
    def __empty(cls) -> Service2TM:
        return cls(subservice=0)

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service2TM:
        service_2_tm = cls.__empty()
        service_2_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        return service_2_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)



"""Contains classes and functions to handle PUS Service 8 telemetry.
"""
from __future__ import annotations
import struct

from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.pus import ObjectId
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class Service8TM(PusTmBase, PusTmInfoBase):
    def __init__(
        self,
        subservice_id: int,
        object_id: bytearray,
        action_id: int,
        custom_data: bytearray,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        pus_tm_version: int = 0b0001,
        ack: int = 0b1111,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        """This class can be used to deserialize service 8 packets.
        :param raw_telemetry:      Raw bytearray which will be deserialized
        :param call_srv8_hook:
        :raises ValueError: If the length of the passed bytearray is too short.
        """
        self._object_id = ObjectId.from_bytes(obj_id_as_bytes=object_id)
        self._action_id = action_id
        self._custom_data = custom_data
        source_data = bytearray()
        source_data.extend(object_id)
        source_data.extend(struct.pack("!I", self._action_id))
        source_data.extend(self._custom_data)
        pus_tm = PusTelemetry(
            service=5,
            subservice=subservice_id,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            pus_tm_version=pus_tm_version,
            ack=ack,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.__init_without_base(instance=self)

    @staticmethod
    def __init_without_base(instance: Service8TM):
        if instance.get_subservice() == 130:
            tm_data = instance.get_tm_data()
            if len(tm_data) < 8:
                LOGGER.warning(
                    f"Length of Service 8 TM data field {len(tm_data)} short than 8"
                )
                raise ValueError
            instance.specify_packet_info("Functional Data Reply")
            instance._object_id = ObjectId.from_bytes(obj_id_as_bytes=tm_data[0:4])
            instance._action_id = struct.unpack("!I", tm_data[4:8])[0]
            instance._custom_data = tm_data[8:]
        else:
            instance.specify_packet_info("Unknown functional commanding reply")

    @classmethod
    def __empty(cls) -> Service8TM:
        return cls(
            subservice_id=-1,
            object_id=bytearray(4),
            action_id=0,
            custom_data=bytearray(),
        )

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ):
        service_8_tm = cls.__empty()
        service_8_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        service_8_tm.__init_without_base(instance=service_8_tm)
        return service_8_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(self._object_id.as_string())
        content_list.append(self._action_id)

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")
        header_list.append("Action ID")

    @property
    def source_object_id_as_bytes(self) -> bytes:
        return self._object_id.as_bytes

    @property
    def source_object_id(self) -> int:
        return self._object_id.as_int

    @property
    def action_id(self) -> int:
        return self._action_id

    @property
    def custom_data(self) -> bytearray:
        return self._custom_data



from abc import abstractmethod
from typing import Dict, Union, Optional, Tuple
from unittest.mock import MagicMock
import argparse

from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.config.com_if import CommunicationInterface
from tmtccmd.config.definitions import DEFAULT_APID
from tmtccmd.config.definitions import ServiceOpCodeDictT, CoreModeList
from tmtccmd.tm.service_3_base import Service3Base
from tmtccmd.core.backend import TmTcHandler
from tmtccmd.tc.definitions import TcQueueT
from tmtccmd.config.hook import TmTcHookBase
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


def create_hook_mock() -> TmTcHookBase:
    """Create simple minimal hook mock using the MagicMock facilities by unittest
    :return:
    """
    tmtc_hook_base = TmTcHookBase()
    tmtc_hook_base.add_globals_pre_args_parsing = MagicMock(return_value=0)
    tmtc_hook_base.add_globals_post_args_parsing = MagicMock(return_value=0)
    tmtc_hook_base.custom_args_parsing = MagicMock(
        return_value=argparse.Namespace(service=17, mode=CoreModeList.IDLE)
    )
    return tmtc_hook_base


def create_hook_mock_with_srv_handlers() -> TmTcHookBase:
    tmtc_hook_base = create_hook_mock()
    tmtc_hook_base.handle_service_8_telemetry = MagicMock(return_value=(["Test"], [0]))
    tmtc_hook_base.handle_service_5_event = MagicMock(return_value="Test Custom String")
    # Valid returnvalue for now
    srv_3_return_tuple = (["Test"], [0], bytearray(0b10000000), 1)
    tmtc_hook_base.handle_service_3_housekeeping = MagicMock(
        return_value=srv_3_return_tuple
    )
    return tmtc_hook_base


class TestHookObj(TmTcHookBase):
    service_8_handler_called = False
    service_5_handler_called = False
    service_3_handler_called = False

    def __init__(self):
        super(self, TmTcHookBase).__init__()
        self.get_obj_id_called = False
        self.add_globals_pre_args_parsing_called = False
        self.add_globals_post_args_parsing_called = False
        self.assign_communication_interface_called = False

    @abstractmethod
    def get_object_ids(self) -> Dict[bytes, list]:
        """The user can specify an object ID dictionary here mapping object ID bytearrays to a
        list. This list could contain containing the string representation or additional
        information about that object ID.
        """
        return TmTcHookBase.get_object_ids()

    @abstractmethod
    def add_globals_pre_args_parsing(self, gui: bool = False):
        """Add all global variables prior to parsing the CLI arguments.

        :param gui: Set to true if the GUI mode is used
        :return:
        """
        from tmtccmd.config.globals import set_default_globals_pre_args_parsing

        set_default_globals_pre_args_parsing(gui=gui, apid=DEFAULT_APID)

    @abstractmethod
    def add_globals_post_args_parsing(self, args: argparse.Namespace):
        """Add global variables prior after parsing the CLI arguments.

        :param args: Specify whether a GUI is used
        """
        from tmtccmd.config.globals import set_default_globals_post_args_parsing

        set_default_globals_post_args_parsing(
            args=args, json_cfg_path=self.get_json_config_file_path()
        )

    @abstractmethod
    def assign_communication_interface(
        self, com_if_key: str, tmtc_printer: TmTcPrinter
    ) -> Optional[CommunicationInterface]:
        """Assign the communication interface used by the TMTC commander to send and receive
        TMTC with.

        :param com_if_key:      String key of the communication interface to be created.
        :param tmtc_printer:    Printer utility instance.
        """
        from tmtccmd.config.com_if import create_communication_interface_default

        return create_communication_interface_default(
            com_if_key=com_if_key,
            tmtc_printer=tmtc_printer,
            json_cfg_path=self.get_json_config_file_path(),
        )

    @abstractmethod
    def get_service_op_code_dictionary(self) -> ServiceOpCodeDictT:
        """This is a dicitonary mapping services represented by strings to an operation code
        dictionary.

        :return:
        """
        from tmtccmd.config.globals import get_default_service_op_code_dict

        return get_default_service_op_code_dict()

    @abstractmethod
    def perform_mode_operation(self, tmtc_backend: TmTcHandler, mode: int):
        """Perform custom mode operations
        :param tmtc_backend:
        :param mode:
        :return:
        """
        pass

    @abstractmethod
    def pack_service_queue(
        self, service: Union[int, str], op_code: str, service_queue: TcQueueT
    ):
        """Overriding this function allows the user to package a telecommand queue for a given
        service and operation code combination.

        :param service:
        :param op_code:
        :param service_queue:
        :return:
        """
        pass

    @staticmethod
    def handle_service_8_telemetry(
        object_id: bytes, action_id: int, custom_data: bytearray
    ) -> Tuple[list, list]:
        """This function is called by the TMTC core to handle Service 8 packets
        The user can return a tuple of two lists, where the first list
        is a list of header strings to print and the second list is a list of values to print.
        The TMTC core will take care of printing both lists and logging them.

        :param object_id: Byte representation of the object ID
        :param action_id:
        :param custom_data:
        :return:
        """
        TestHookObj.service_8_handler_called = True
        return [], []

    @staticmethod
    def handle_service_3_housekeeping(
        object_id: bytes, set_id: int, hk_data: bytearray, service3_packet: Service3Base
    ) -> Tuple[list, list, bytearray, int]:
        """This function is called when a Service 3 Housekeeping packet is received.

        :param object_id: Byte representation of the object ID
        :param set_id: Unique set ID of the HK reply
        :param hk_data:     HK data. For custom HK handling, whole HK data will be passed here.
                            Otherwise, a 8 byte SID consisting of the 4 byte object ID and 4 byte
                            set ID will be assumed and the remaining packet after the first 4 bytes
                            will be passed here.
        :param service3_packet: Service 3 packet object
        :return: Expects a tuple, consisting of two lists, a bytearray and an integer
            The first list contains the header columns, the second list the list with
            the corresponding values. The bytearray is the validity buffer, which is usually appended
            at the end of the housekeeping packet. The last value is the number of parameters.
        """
        LOGGER.info(
            "TmTcHookBase: No service 3 housekeeping data handling implemented yet in "
            "handle_service_3_housekeeping hook function"
        )
        return [], [], bytearray(), 0

    @staticmethod
    def handle_service_5_event(
        object_id: bytes, event_id: int, param_1: int, param_2: int
    ) -> str:
        """This function is called when a Service 5 Event Packet is received. The user can specify
         a custom string here which will be printed to display additional information related
         to an event.

        :param object_id: Byte representation of the object ID
        :param event_id: Two-byte event ID
        :param param_1: Four-byte Parameter 1
        :param param_2: Four-byte Parameter 2
        :return: Custom information string which will be printed with the event
        """
        return ""



"""Contains definitions and functions related to PUS Service 20 Telecommands.
"""
from typing import Union
from spacepackets.ecss.conf import get_default_tc_apid
from spacepackets.ecss.tc import PusTelecommand
from tmtccmd.pus.service_20_parameter import EcssPtc, EcssPfcUnsigned
from tmtccmd.utility.logger import get_console_logger

logger = get_console_logger()


def pack_boolean_parameter_command(
    object_id: bytearray,
    domain_id: int,
    unique_id: int,
    parameter: bool,
    ssc: int,
    apid: int = -1,
) -> Union[PusTelecommand, None]:
    """Generic function to pack a telecommand to tweak a boolean parameter
    :param object_id:
    :param domain_id:
    :param unique_id:
    :param parameter:
    :param ssc:
    :param apid:
    @return:
    """
    if apid == -1:
        apid = get_default_tc_apid()

    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    if unique_id > 255:
        logger.warning("Invalid unique ID, should be smaller than 255!")
        return None
    parameter_id[1] = unique_id
    parameter_id[2] = 0
    parameter_id[3] = 0
    data_to_pack = bytearray(object_id)
    data_to_pack.extend(parameter_id)
    # PTC and PFC for uint8_t according to CCSDS
    ptc = EcssPtc.UNSIGNED
    pfc = EcssPfcUnsigned.ONE_BYTE
    rows = 1
    columns = 1
    data_to_pack.append(ptc)
    data_to_pack.append(pfc)
    data_to_pack.append(rows)
    data_to_pack.append(columns)
    data_to_pack.append(parameter)
    return PusTelecommand(
        service=20, subservice=128, ssc=ssc, app_data=data_to_pack, apid=apid
    )


def pack_float_vector_parameter_command(
    object_id: bytearray,
    domain_id: int,
    unique_id: int,
    parameter: bytearray,
    ssc: int,
    apid: int = -1,
):
    pass


def pack_type_and_matrix_data(ptc: int, pfc: int, rows: int, columns: int) -> bytearray:
    # noinspection PyPep8
    """Packs the parameter information field, which contains the ECSS PTC and PFC numbers and the number of columns
    and rows in the parameter.
    See https://ecss.nl/standard/ecss-e-st-70-41c-space-engineering-telemetry-and-telecommand-packet-utilization-15-april-2016/
    p.428 for more information.
    :param ptc:     ECSS PTC number
    :param pfc:     ECSS PFC number
    :param rows:     Number of rows in parameter (for matrix entries, 1 for vector entries, 1 for scalar entries)
    :param columns:  Number of columns in parameter (for matrix or vector entries, 1 for scalar entries)
    :return: Parameter information field as 4 byte bytearray
    """
    data = bytearray(4)
    data[0] = ptc
    data[1] = pfc
    data[2] = rows
    data[3] = columns
    return data


def pack_parameter_id(domain_id: int, unique_id: int, linear_index: int) -> bytearray:
    """Packs the Parameter ID (bytearray with 4 bytes) which is part of the service 20 packets.
    The first byte of the parameter ID is the domain ID, the second byte is a unique ID and the
    last two bytes are a linear index if a parameter is not loaded from index 0.
    :param domain_id:       One byte domain ID
    :param unique_id:       One byte unique ID
    :param linear_index:    Two byte linear index.
    """
    parameter_id = bytearray(4)
    parameter_id[0] = domain_id
    parameter_id[1] = unique_id
    parameter_id[2] = linear_index >> 8 & 0xFF
    parameter_id[3] = linear_index & 0xFF
    return parameter_id



"""Definitions for the TMTC commander core
"""
import enum
from typing import Tuple, Dict, Optional, List, Union


class CoreGlobalIds(enum.IntEnum):
    """
    Numbers from 128 to 200 are reserved for core globals
    """

    # Object handles
    TMTC_HOOK = 128
    COM_INTERFACE_HANDLE = 129
    TM_LISTENER_HANDLE = 130
    TMTC_PRINTER_HANDLE = 131
    TM_HANDLER_HANDLE = 132
    PRETTY_PRINTER = 133

    # Parameters
    JSON_CFG_PATH = 139
    MODE = 141
    CURRENT_SERVICE = 142
    COM_IF = 144
    OP_CODE = 145
    TM_TIMEOUT = 146
    SERVICE_OP_CODE_DICT = 147
    COM_IF_DICT = 148

    # Miscellaneous
    DISPLAY_MODE = 150
    USE_LISTENER_AFTER_OP = 151
    PRINT_HK = 152
    PRINT_TM = 153
    PRINT_RAW_TM = 154
    PRINT_TO_FILE = 155
    RESEND_TC = 156
    TC_SEND_TIMEOUT_FACTOR = 157

    # Config dictionaries
    USE_SERIAL = 160
    SERIAL_CONFIG = 161
    USE_ETHERNET = 162
    ETHERNET_CONFIG = 163


class OpCodeDictKeys(enum.IntEnum):
    TIMEOUT = CoreGlobalIds.TM_TIMEOUT


# Service Op Code Dictionary Types
ServiceNameT = str
ServiceInfoT = str
OpCodeNameT = Union[str, List[str]]
OpCodeInfoT = str
# Operation code options are optional. If none are supplied, default values are assumed
OpCodeOptionsT = Optional[Dict[OpCodeDictKeys, any]]
OpCodeEntryT = Dict[OpCodeNameT, Tuple[OpCodeInfoT, OpCodeOptionsT]]
# It is possible to specify a service without any op codes
ServiceDictValueT = Optional[Tuple[ServiceInfoT, OpCodeEntryT]]
ServiceOpCodeDictT = Dict[ServiceNameT, ServiceDictValueT]

# Com Interface Types
ComIFValueT = Tuple[str, any]
ComIFDictT = Dict[str, ComIFValueT]

EthernetAddressT = Tuple[str, int]


class CoreComInterfaces(enum.Enum):
    DUMMY = "dummy"
    SERIAL_DLE = "ser_dle"
    TCPIP_UDP = "udp"
    TCPIP_TCP = "tcp"
    SERIAL_FIXED_FRAME = "ser_fixed"
    SERIAL_QEMU = "ser_qemu"
    UNSPECIFIED = "unspec"


CoreComInterfacesDict = {
    CoreComInterfaces.DUMMY.value: "Dummy Interface",
    CoreComInterfaces.SERIAL_DLE.value: "Serial Interace with DLE encoding",
    CoreComInterfaces.TCPIP_UDP.value: "TCP/IP with UDP datagrams",
    CoreComInterfaces.TCPIP_TCP.value: "TCP/IP with TCP",
    CoreComInterfaces.SERIAL_FIXED_FRAME.value: "Serial Interface with fixed size frames",
    CoreComInterfaces.SERIAL_QEMU.value: "Serial Interface using QEMU",
    CoreComInterfaces.UNSPECIFIED.value: "Unspecified",
}


class QueueCommands(enum.Enum):
    PRINT = enum.auto()
    RAW_PRINT = enum.auto()
    WAIT = enum.auto()
    EXPORT_LOG = enum.auto()
    SET_TIMEOUT = enum.auto()


# Mode options, set by args parser
class CoreModeList(enum.IntEnum):
    SEQUENTIAL_CMD_MODE = 0
    LISTENER_MODE = 1
    GUI_MODE = 2
    IDLE = 5
    PROMPT_MODE = 6


CoreModeStrings = {
    CoreModeList.SEQUENTIAL_CMD_MODE: "seqcmd",
    CoreModeList.LISTENER_MODE: "listener",
    CoreModeList.GUI_MODE: "gui",
}


class CoreServiceList(enum.Enum):
    SERVICE_2 = "2"
    SERVICE_3 = "3"
    SERVICE_5 = "5"
    SERVICE_8 = "8"
    SERVICE_9 = "9"
    SERVICE_11 = "11"
    SERVICE_17 = "17"
    SERVICE_20 = "20"
    SERVICE_23 = "23"
    SERVICE_200 = "200"


DEFAULT_APID = 0xEF
DEBUG_MODE = False



import enum


class Srv5Severity(enum.IntEnum):
    INFO = 1
    LOW_SEVERITY = 2
    MEDIUM_SEVERITY = 3
    HIGH_SEVERITY = 4


class Srv5Subservices(enum.IntEnum):
    INFO_EVENT = (Srv5Severity.INFO,)
    LOW_SEVERITY_EVENT = (Srv5Severity.LOW_SEVERITY,)
    MEDIUM_SEVERITY_EVENT = (Srv5Severity.MEDIUM_SEVERITY,)
    HIGH_SEVERITY_EVENT = (Srv5Severity.HIGH_SEVERITY,)
    ENABLE_EVENT_REPORTING = (5,)
    DISABLE_EVENT_REPORTING = 6



from __future__ import annotations
import os
import struct

from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry
from spacepackets.ecss.definitions import PusServices

from tmtccmd.pus.obj_id import ObjectId
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class ParamStruct:
    def __init__(self):
        self.param_id = (0,)
        self.domain_id = (0,)
        self.unique_id = (0,)
        self.linear_index = (0,)
        self.type_ptc = 0
        self.type_pfc = 0
        self.column = (0,)
        self.row = (0,)
        self.param: any = 0


class Service20TM(PusTmInfoBase, PusTmBase):
    def __init__(
        self,
        subservice_id: int,
        object_id: bytearray,
        param_id: bytearray,
        domain_id: int,
        unique_id: int,
        linear_index: int,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        pus_tm = PusTelemetry(
            service=PusServices.SERVICE_20_PARAMETER,
            subservice=subservice_id,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )

        self.object_id = ObjectId.from_bytes(obj_id_as_bytes=object_id)
        self.param_struct = ParamStruct()
        self.param_struct.param_id = param_id
        self.param_struct.domain_id = domain_id
        self.param_struct.unique_id = unique_id
        self.param_struct.linear_index = linear_index
        PusTmBase.__init__(self, pus_tm=pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=pus_tm)
        self.__init_without_base(instance=self)
        self.set_packet_info("Parameter Service Reply")

    @staticmethod
    def __init_without_base(instance: Service20TM):
        tm_data = instance.tm_data
        if len(tm_data) < 8:
            return
        data_size = len(tm_data)
        instance.object_id = ObjectId.from_bytes(obj_id_as_bytes=tm_data[0:4])
        instance.param_struct.param_id = struct.unpack("!I", tm_data[4:8])[0]
        instance.param_struct.domain_id = tm_data[4]
        instance.param_struct.unique_id = tm_data[5]
        instance.param_struct.linear_index = tm_data[6] << 8 | tm_data[7]

        if instance.subservice == 130:
            # TODO: This needs to be more generic. Furthermore, we need to be able to handle
            #       vector and matrix dumps as well and this is not possible in the current form.
            instance.param_struct.type_ptc = tm_data[8]
            instance.param_struct.type_pfc = tm_data[9]
            instance.param_struct.column = tm_data[10]
            instance.param_struct.row = tm_data[11]
            if data_size > 12:
                if (
                    instance.param_struct.type_ptc == 3
                    and instance.param_struct.type_pfc == 14
                ):
                    instance.param_struct.param = struct.unpack("!I", tm_data[12:16])[0]
                if (
                    instance.param_struct.type_ptc == 4
                    and instance.param_struct.type_pfc == 14
                ):
                    instance.param_struct.param = struct.unpack("!i", tm_data[12:16])[0]
                if (
                    instance.param_struct.type_ptc == 5
                    and instance.param_struct.type_pfc == 1
                ):
                    instance.param_struct.param = struct.unpack("!f", tm_data[12:16])[0]
            else:
                LOGGER.info(
                    "Error when receiving Pus Service 20 TM: subservice is not 130"
                )

    @classmethod
    def __empty(cls) -> Service20TM:
        return cls(
            subservice_id=-1,
            object_id=bytearray(4),
            param_id=bytearray(),
            domain_id=0,
            unique_id=0,
            linear_index=0,
        )

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service20TM:
        service_20_tm = cls.__empty()
        service_20_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        if len(service_20_tm.pus_tm.tm_data) < 4:
            LOGGER.warning("Invalid data length, less than 4")
        elif len(service_20_tm.pus_tm.tm_data) < 8:
            LOGGER.warning(
                "Invalid data length, less than 8 (Object ID and Parameter ID)"
            )
        service_20_tm.__init_without_base(instance=service_20_tm)
        return service_20_tm

    def append_telemetry_content(self, content_list: list):
        super().append_telemetry_content(content_list=content_list)
        content_list.append(self.object_id.as_string)

    def append_telemetry_column_headers(self, header_list: list):
        super().append_telemetry_column_headers(header_list=header_list)
        header_list.append("Object ID")

    def get_custom_printout(self) -> str:
        custom_printout = ""
        header_list = []
        content_list = []
        if self.subservice == 130:
            custom_printout = f"Parameter Information:{os.linesep}"
            header_list.append("Domain ID")
            header_list.append("Unique ID")
            header_list.append("Linear Index")
            header_list.append("CCSDS Type")
            header_list.append("Columns")
            header_list.append("Rows")
            # TODO: For more complex parameters like vectors or matrices,
            #       special handling would be nice
            header_list.append("Parameter")

            content_list.append(self.param_struct.domain_id)
            content_list.append(self.param_struct.unique_id)
            content_list.append(self.param_struct.linear_index)
            content_list.append(
                f"PTC: {self.param_struct.type_ptc} | PFC: {self.param_struct.type_pfc}"
            )
            content_list.append(self.param_struct.column)
            content_list.append(self.param_struct.row)
            content_list.append(self.param_struct.param)

            custom_printout += f"{header_list}{os.linesep}"
            custom_printout += f"{content_list}"
        return custom_printout



from __future__ import annotations
import enum
import sys

from spacepackets.ecss.tm import CdsShortTimestamp, PusVersion, PusTelemetry
from spacepackets.ecss.service_17_test import Service17TM
from spacepackets.ecss.definitions import PusServices
from spacepackets.ecss.conf import get_default_tc_apid

from tmtccmd.config.definitions import QueueCommands
from tmtccmd.tc.definitions import PusTelecommand, TcQueueT
from tmtccmd.tm.base import PusTmInfoBase, PusTmBase


class Srv17Subservices(enum.IntEnum):
    PING_CMD = (1,)
    PING_REPLY = (2,)
    GEN_EVENT = 128


class Service17TMExtended(PusTmBase, PusTmInfoBase, Service17TM):
    def __init__(
        self,
        subservice: int,
        time: CdsShortTimestamp = None,
        ssc: int = 0,
        source_data: bytearray = bytearray([]),
        apid: int = -1,
        packet_version: int = 0b000,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
        secondary_header_flag: bool = True,
        space_time_ref: int = 0b0000,
        destination_id: int = 0,
    ):
        Service17TM.__init__(
            self,
            subservice=subservice,
            time=time,
            ssc=ssc,
            source_data=source_data,
            apid=apid,
            packet_version=packet_version,
            pus_version=pus_version,
            secondary_header_flag=secondary_header_flag,
            space_time_ref=space_time_ref,
            destination_id=destination_id,
        )
        PusTmBase.__init__(self, pus_tm=self.pus_tm)
        PusTmInfoBase.__init__(self, pus_tm=self.pus_tm)
        if self.subservice == Srv17Subservices.PING_REPLY:
            self.set_packet_info("Ping Reply")

    @classmethod
    def __empty(cls) -> Service17TMExtended:
        return cls(subservice=0)

    @classmethod
    def unpack(
        cls,
        raw_telemetry: bytearray,
        pus_version: PusVersion = PusVersion.GLOBAL_CONFIG,
    ) -> Service17TMExtended:
        service_17_tm = cls.__empty()
        service_17_tm.pus_tm = PusTelemetry.unpack(
            raw_telemetry=raw_telemetry, pus_version=pus_version
        )
        return service_17_tm


def pack_service_17_ping_command(ssc: int, apid: int = -1) -> PusTelecommand:
    """Generate a simple ping PUS telecommand packet"""
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=17, subservice=Srv17Subservices.PING_CMD, ssc=ssc, apid=apid
    )


def pack_generic_service17_test(
    init_ssc: int, tc_queue: TcQueueT, apid: int = -1
) -> int:
    if apid == -1:
        apid = get_default_tc_apid()
    new_ssc = init_ssc
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 17"))
    # ping test
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 17: Ping Test"))
    tc_queue.appendleft(pack_service_17_ping_command(ssc=new_ssc).pack_command_tuple())
    new_ssc += 1
    # enable event
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 17: Enable Event"))
    command = PusTelecommand(service=5, subservice=5, ssc=new_ssc, apid=apid)
    tc_queue.appendleft(command.pack_command_tuple())
    new_ssc += 1
    # test event
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 17: Trigger event"))
    command = PusTelecommand(
        service=17, subservice=Srv17Subservices.GEN_EVENT, ssc=new_ssc, apid=apid
    )
    tc_queue.appendleft(command.pack_command_tuple())
    new_ssc += 1
    # invalid subservice
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 17: Invalid subservice"))
    command = PusTelecommand(service=17, subservice=243, ssc=new_ssc, apid=apid)
    tc_queue.appendleft(command.pack_command_tuple())
    new_ssc += 1
    return new_ssc



"""Contains definitions and functions related to PUS Service 5 Telecommands.
"""
from spacepackets.ecss.conf import get_default_tc_apid

from tmtccmd.config.definitions import QueueCommands
from tmtccmd.tc.definitions import PusTelecommand, TcQueueT
from tmtccmd.pus.service_5_event import Srv5Subservices


def pack_enable_event_reporting_command(ssc: int, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=5, subservice=Srv5Subservices.ENABLE_EVENT_REPORTING, ssc=ssc, apid=apid
    )


def pack_disable_event_reporting_command(ssc: int, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    return PusTelecommand(
        service=5,
        subservice=Srv5Subservices.DISABLE_EVENT_REPORTING,
        ssc=ssc,
        apid=apid,
    )


def pack_generic_service5_test_into(tc_queue: TcQueueT, apid: int = -1):
    if apid == -1:
        apid = get_default_tc_apid()
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 5"))
    # invalid subservice
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 5: Invalid subservice"))
    command = PusTelecommand(service=5, subservice=1, apid=apid, ssc=500)
    tc_queue.appendleft(command.pack_command_tuple())
    # disable events
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 5: Disable event"))
    command = pack_disable_event_reporting_command(ssc=501)
    tc_queue.appendleft(command.pack_command_tuple())
    # trigger event
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 5: Trigger event"))
    command = PusTelecommand(service=17, subservice=128, apid=apid, ssc=510)
    tc_queue.appendleft(command.pack_command_tuple())
    # enable event
    tc_queue.appendleft((QueueCommands.PRINT, "Testing Service 5: Enable event"))
    command = pack_enable_event_reporting_command(ssc=520)
    tc_queue.appendleft(command.pack_command_tuple())
    # trigger event
    tc_queue.appendleft(
        (QueueCommands.PRINT, "Testing Service 5: Trigger another event")
    )
    command = PusTelecommand(service=17, subservice=128, apid=apid, ssc=530)
    tc_queue.appendleft(command.pack_command_tuple())
    tc_queue.appendleft((QueueCommands.EXPORT_LOG, "log/tmtc_log_service5.txt"))



#!/usr/bin/env python3
from unittest import TestCase

from spacepackets.ccsds.spacepacket import get_space_packet_sequence_control
from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.conf import get_pus_tm_version, PusVersion, set_default_tm_apid
from spacepackets.util import PrintFormats

from tmtccmd.pus.service_17_test import Service17TMExtended


class TestTelemetry(TestCase):
    def test_generic_pus_c(self):
        pus_17_telemetry = Service17TMExtended(
            subservice=1,
            ssc=36,
            time=CdsShortTimestamp.init_from_current_time(),
            apid=0xEF,
        )
        pus_17_raw = pus_17_telemetry.pack()

        pus_17_telemetry = None

        def tm_func(raw_telemetry: bytearray):
            return Service17TMExtended.unpack(raw_telemetry=raw_telemetry)

        self.assertRaises(ValueError, tm_func, bytearray())
        self.assertRaises(ValueError, tm_func, None)

        pus_17_telemetry = Service17TMExtended.unpack(raw_telemetry=pus_17_raw)
        self.assertTrue(get_pus_tm_version() == PusVersion.PUS_C)
        self.assertTrue(pus_17_telemetry.service == 17)
        self.assertTrue(pus_17_telemetry.apid == 0xEF)
        self.assertTrue(pus_17_telemetry.subservice == 1)
        self.assertTrue(pus_17_telemetry.ssc == 36)
        self.assertTrue(pus_17_telemetry.tm_data == bytearray())
        self.assertTrue(pus_17_telemetry.pus_tm.valid)
        self.assertTrue(pus_17_telemetry.get_custom_printout() == "")
        self.assertTrue(pus_17_telemetry.get_source_data_string() == "hex []")
        pus_17_telemetry.pus_tm.print_source_data(print_format=PrintFormats.HEX)
        pus_17_telemetry.pus_tm.print_full_packet_string(print_format=PrintFormats.HEX)
        # This string changes depending on system time, so its complicated to test its validity
        full_string = pus_17_telemetry.pus_tm.get_full_packet_string(
            print_format=PrintFormats.HEX
        )
        print(full_string)
        print(pus_17_telemetry)
        print(repr(pus_17_telemetry))
        self.assertTrue(pus_17_telemetry.pus_tm.packet_id == 0x8 << 8 | 0xEF)

    def test_list_functionality(self):
        pus_17_telecommand = Service17TMExtended(
            subservice=1, ssc=36, time=CdsShortTimestamp.init_from_current_time()
        )
        pus_17_raw = pus_17_telecommand.pack()
        pus_17_telemetry = Service17TMExtended.unpack(raw_telemetry=pus_17_raw)

        header_list = []
        content_list = []
        pus_17_telemetry.append_telemetry_column_headers(header_list=header_list)
        pus_17_telemetry.append_telemetry_content(content_list=content_list)

        self.assertTrue(header_list != [])
        self.assertTrue(content_list != [])



#!/usr/bin/python3
"""
We do the package handling in the static setup.cfg but include an empty setup.py
to allow editable installs https://packaging.python.org/tutorials/packaging-projects/
and provide extensibility
"""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup()



# -*- coding: utf-8 -*-
"""
:file:      obsw_tc_packer.py
:author:    R. Mueller
:date:      10.05.2021
"""
import sys
from typing import Union

from tmtccmd.tc.definitions import TcQueueT
from spacepackets.ecss.tc import PusTelecommand
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.pus.service_17_test import pack_service_17_ping_command
from tmtccmd.tc.service_5_event import pack_generic_service5_test_into

LOGGER = get_console_logger()


class ServiceQueuePacker:
    def __init__(self):
        pass

    @staticmethod
    def pack_service_queue_core(service: int, op_code: str, service_queue: TcQueueT):
        """
        Use hook object supplied by user
        """
        try:
            from tmtccmd.config.hook import get_global_hook_obj

            hook_obj = get_global_hook_obj()
            hook_obj.pack_service_queue(
                service=service, op_code=op_code, service_queue=service_queue
            )
        except ImportError:
            LOGGER.exception(
                "Could not import custom telecommand hook! Make sure to implement it."
            )
            sys.exit(1)


def default_single_packet_preparation() -> PusTelecommand:
    return pack_service_17_ping_command(ssc=1700)


def default_service_queue_preparation(
    service: Union[str, int], op_code: str, service_queue: TcQueueT
):
    from tmtccmd.config.definitions import CoreServiceList, QueueCommands

    if service == CoreServiceList.SERVICE_5.value:
        pack_generic_service5_test_into(service_queue)
    elif service == CoreServiceList.SERVICE_17.value:
        service_queue.appendleft(
            (QueueCommands.PRINT, "Sending ping command PUS TC[17,1]")
        )
        service_queue.appendleft(
            pack_service_17_ping_command(ssc=1700).pack_command_tuple()
        )
    else:
        LOGGER.warning("Invalid Service!")



#!/usr/bin/python3.8
"""
@file
    tmtcc_config.py
@date
    01.11.2019
@brief
    Used to send single tcs and listen for replies after that
"""
from tmtccmd.ccsds.handler import CcsdsTmHandler
from tmtccmd.sendreceive.cmd_sender_receiver import CommandSenderReceiver
from tmtccmd.sendreceive.tm_listener import TmListener

from tmtccmd.com_if.com_interface_base import CommunicationInterface

from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.utility.logger import get_console_logger

from tmtccmd.tc.definitions import PusTcTupleT


logger = get_console_logger()


class SingleCommandSenderReceiver(CommandSenderReceiver):
    """
    Specific implementation of CommandSenderReceiver to send a single telecommand
    This object can be used by instantiating it and calling sendSingleTcAndReceiveTm()
    """

    def __init__(
        self,
        com_if: CommunicationInterface,
        tmtc_printer: TmTcPrinter,
        tm_listener: TmListener,
        tm_handler: CcsdsTmHandler,
        apid: int,
    ):
        """
        :param com_if: CommunicationInterface object, passed on to CommandSenderReceiver
        :param tm_listener: TmListener object which runs in the background and receives all TM
        :param tmtc_printer: TmTcPrinter object, passed on to CommandSenderReceiver
        """
        super().__init__(
            com_if=com_if,
            tm_listener=tm_listener,
            tmtc_printer=tmtc_printer,
            tm_handler=tm_handler,
            apid=apid,
        )

    def send_single_tc_and_receive_tm(self, pus_packet_tuple: PusTcTupleT):
        """
        Send a single telecommand passed to the class and wait for replies
        :return:
        """
        try:
            pus_packet_raw, pus_packet_obj = pus_packet_tuple
        except TypeError:
            logger.error("SingleCommandSenderReceiver: Invalid command input")
            return
        self._operation_pending = True
        self._tm_listener.set_listener_mode(TmListener.ListenerModes.SEQUENCE)
        self._tmtc_printer.print_telecommand(
            tc_packet_obj=pus_packet_obj, tc_packet_raw=pus_packet_raw
        )
        self._com_if.send(data=pus_packet_raw)
        self._last_tc = pus_packet_raw
        self._last_tc_obj = pus_packet_obj
        while self._operation_pending:
            # wait until reply is received
            super()._check_for_first_reply()
        if self._reply_received:
            self._tm_listener.set_mode_op_finished()
            packet_queue = self._tm_listener.retrieve_ccsds_tm_packet_queue(
                apid=self._apid, clear=True
            )
            self._tm_handler.handle_ccsds_packet_queue(
                apid=self._apid, packet_queue=packet_queue
            )
            logger.info("SingleCommandSenderReceiver: Reply received")
            logger.info("Listening for packages ...")



import sys
import argparse
from abc import abstractmethod
from typing import Dict, Tuple, Optional, Union

from tmtccmd.config.definitions import DEFAULT_APID, ServiceOpCodeDictT
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.core.backend import TmTcHandler
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.tc.definitions import TcQueueT
from tmtccmd.com_if.com_interface_base import CommunicationInterface
from tmtccmd.tm.service_3_base import Service3Base

LOGGER = get_console_logger()


class TmTcHookBase:
    """This hook allows users to adapt the TMTC commander core to the unique mission requirements.
    It is used by implementing all abstract functions and then passing the instance to the
    TMTC commander core.
    """

    def __init__(self):
        pass

    @abstractmethod
    def get_object_ids(self) -> Dict[bytes, list]:
        from tmtccmd.config.objects import get_core_object_ids

        """The user can specify an object ID dictionary here mapping object ID bytearrays to a
        list. This list could contain containing the string representation or additional
        information about that object ID.
        """
        return get_core_object_ids()

    @abstractmethod
    def add_globals_pre_args_parsing(self, gui: bool = False):
        """Add all global variables prior to parsing the CLI arguments.

        :param gui: Set to true if the GUI mode is used
        :return:
        """
        from tmtccmd.config.globals import set_default_globals_pre_args_parsing

        set_default_globals_pre_args_parsing(gui=gui, apid=DEFAULT_APID)

    @abstractmethod
    def add_globals_post_args_parsing(self, args: argparse.Namespace):
        """Add global variables prior after parsing the CLI arguments.

        :param args: Specify whether a GUI is used
        """
        from tmtccmd.config.globals import set_default_globals_post_args_parsing

        set_default_globals_post_args_parsing(
            args=args, json_cfg_path=self.get_json_config_file_path()
        )

    @abstractmethod
    def assign_communication_interface(
        self, com_if_key: str, tmtc_printer: TmTcPrinter
    ) -> Optional[CommunicationInterface]:
        """Assign the communication interface used by the TMTC commander to send and receive
        TMTC with.

        :param com_if_key:      String key of the communication interface to be created.
        :param tmtc_printer:    Printer utility instance.
        """
        from tmtccmd.config.com_if import create_communication_interface_default

        return create_communication_interface_default(
            com_if_key=com_if_key,
            tmtc_printer=tmtc_printer,
            json_cfg_path=self.get_json_config_file_path(),
        )

    @abstractmethod
    def get_service_op_code_dictionary(self) -> ServiceOpCodeDictT:
        """This is a dicitonary mapping services represented by strings to an operation code
        dictionary.

        :return:
        """
        from tmtccmd.config.globals import get_default_service_op_code_dict

        return get_default_service_op_code_dict()

    @abstractmethod
    def perform_mode_operation(self, tmtc_backend: TmTcHandler, mode: int):
        """Perform custom mode operations
        :param tmtc_backend:
        :param mode:
        :return:
        """
        pass

    @abstractmethod
    def pack_service_queue(
        self, service: Union[int, str], op_code: str, service_queue: TcQueueT
    ):
        """Overriding this function allows the user to package a telecommand queue for a given
        service and operation code combination.

        :param service:
        :param op_code:
        :param service_queue:
        :return:
        """
        pass

    @staticmethod
    def custom_args_parsing() -> Optional[argparse.Namespace]:
        """The user can implement args parsing here to override the default argument parsing
        for the CLI mode
        :return:
        """
        return None

    def get_json_config_file_path(self) -> str:
        """The user can specify a path and filename for the JSON configuration file by overriding
        this function.

        :return:
        """
        return "tmtc_config.json"

    @staticmethod
    def handle_service_8_telemetry(
        object_id: bytes, action_id: int, custom_data: bytearray
    ) -> Tuple[list, list]:
        """This function is called by the TMTC core to handle Service 8 packets
        The user can return a tuple of two lists, where the first list
        is a list of header strings to print and the second list is a list of values to print.
        The TMTC core will take care of printing both lists and logging them.

        :param object_id: Byte representation of the object ID
        :param action_id:
        :param custom_data:
        :return:
        """
        LOGGER.info(
            "TmTcHookBase: No service 8 handling implemented yet in handle_service_8_telemetry "
            "hook function"
        )
        return [], []

    @staticmethod
    def handle_service_3_housekeeping(
        object_id: bytes, set_id: int, hk_data: bytearray, service3_packet: Service3Base
    ) -> Tuple[list, list, bytearray, int]:
        """This function is called when a Service 3 Housekeeping packet is received.

        :param object_id: Byte representation of the object ID
        :param set_id: Unique set ID of the HK reply
        :param hk_data:     HK data. For custom HK handling, whole HK data will be passed here.
                            Otherwise, a 8 byte SID consisting of the 4 byte object ID and 4 byte
                            set ID will be assumed and the remaining packet after the first 4 bytes
                            will be passed here.
        :param service3_packet: Service 3 packet object
        :return: Expects a tuple, consisting of two lists, a bytearray and an integer
            The first list contains the header columns, the second list the list with
            the corresponding values. The bytearray is the validity buffer, which is usually appended
            at the end of the housekeeping packet. The last value is the number of parameters.
        """
        LOGGER.info(
            "TmTcHookBase: No service 3 housekeeping data handling implemented yet in "
            "handle_service_3_housekeeping hook function"
        )
        return [], [], bytearray(), 0

    @staticmethod
    def handle_service_5_event(
        object_id: bytes, event_id: int, param_1: int, param_2: int
    ) -> str:
        """This function is called when a Service 5 Event Packet is received. The user can specify
         a custom string here which will be printed to display additional information related
         to an event.

        :param object_id: Byte representation of the object ID
        :param event_id: Two-byte event ID
        :param param_1: Four-byte Parameter 1
        :param param_2: Four-byte Parameter 2
        :return: Custom information string which will be printed with the event
        """
        return ""


def get_global_hook_obj() -> Optional[TmTcHookBase]:
    """This function can be used to get the handle to the global hook object.
    :return:
    """
    try:
        from tmtccmd.core.globals_manager import get_global
        from tmtccmd.config.definitions import CoreGlobalIds

        from typing import cast

        hook_obj_raw = get_global(CoreGlobalIds.TMTC_HOOK)
        if hook_obj_raw is None:
            LOGGER.error("Hook object is invalid!")
            sys.exit(0)
        return cast(TmTcHookBase, hook_obj_raw)
    except ImportError:
        LOGGER.exception("Issues importing modules to get global hook handle!")
        return None
    except AttributeError:
        LOGGER.exception("Attribute error when trying to get global hook handle!")
        return None



"""Contains classes and functions that perform all printing functionalities.
"""
import os
import enum
from typing import cast

from spacepackets.ecss.tc import PusTelecommand
from spacepackets.util import get_printable_data_string, PrintFormats

from tmtccmd.tm.service_8_functional_cmd import Service8TM
from tmtccmd.tm.service_5_event import Service5TM
from spacepackets.ecss.definitions import PusServices
from tmtccmd.tm.base import PusTmInfoInterface, PusTmInterface
from tmtccmd.pus.service_8_func_cmd import Srv8Subservices
from tmtccmd.tm.definitions import PusIFQueueT
from tmtccmd.tm.service_3_base import Service3Base, HkContentType
from tmtccmd.utility.logger import get_console_logger

LOGGER = get_console_logger()


class DisplayMode(enum.Enum):
    """List of display modes"""

    SHORT = enum.auto()
    LONG = enum.auto()


class TmTcPrinter:
    """This class handles printing to the command line and to files.
    TODO: Introduce file logger"""

    def __init__(
        self,
        display_mode: DisplayMode = DisplayMode.LONG,
        do_print_to_file: bool = True,
        print_tc: bool = True,
    ):
        """
        :param display_mode:
        :param do_print_to_file: if true, print to file
        :param print_tc: if true, print TCs
        """
        self._display_mode = display_mode
        self.do_print_to_file = do_print_to_file
        self.print_tc = print_tc
        self.__print_buffer = ""
        # global print buffer which will be useful to print something to file
        self.__file_buffer = ""
        # List implementation to store multiple strings
        self.file_buffer_list = []

    def set_display_mode(self, display_mode: DisplayMode):
        self._display_mode = display_mode

    def get_display_mode(self) -> DisplayMode:
        return self._display_mode

    def print_telemetry_queue(self, tm_queue: PusIFQueueT):
        """Print the telemetry queue which should contain lists of TM class instances."""
        for tm_list in tm_queue:
            for tm_packet in tm_list:
                self.print_telemetry(packet_if=tm_packet, info_if=tm_packet)

    def print_telemetry(
        self,
        packet_if: PusTmInterface,
        info_if: PusTmInfoInterface,
        print_raw_tm: bool = False,
    ):
        """This function handles printing telemetry
        :param packet_if:       Core interface to work with PUS packets
        :param info_if:         Core interface to get custom data from PUS packets
        :param print_raw_tm:    Specify whether the TM packet is printed in a raw way.
        :return:
        """
        if not isinstance(packet_if, PusTmInterface) or not isinstance(
            info_if, PusTmInfoInterface
        ):
            LOGGER.warning("Passed packet does not implement necessary interfaces!")
            return

        if packet_if.service == PusServices.SERVICE_5_EVENT:
            self.__handle_event_packet(cast(Service5TM, packet_if))

        if self._display_mode == DisplayMode.SHORT:
            self.__handle_short_print(packet_if)
        else:
            self.__handle_long_tm_print(packet_if=packet_if, info_if=info_if)
        self.__handle_wiretapping_packet(packet_if=packet_if, info_if=info_if)

        # Handle special packet types
        if packet_if.service == PusServices.SERVICE_3_HOUSEKEEPING:
            self.handle_service_3_packet(packet_if=packet_if)
        if packet_if.service == PusServices.SERVICE_5_EVENT:
            self.handle_service_5_packet(packet_if=packet_if)
        if (
            packet_if.service == PusServices.SERVICE_8_FUNC_CMD
            and packet_if.subservice == Srv8Subservices.DATA_REPLY
        ):
            self.handle_service_8_packet(packet_if=packet_if)

        if print_raw_tm:
            tm_data_string = get_printable_data_string(
                print_format=PrintFormats.HEX, data=packet_if.pack()
            )
            self.__print_buffer = f"TM Data:\n{tm_data_string}"
            LOGGER.info(self.__print_buffer)
            self.add_print_buffer_to_file_buffer()

    def handle_service_3_packet(self, packet_if: PusTmInterface):
        from tmtccmd.config.hook import get_global_hook_obj

        if packet_if.service != 3:
            LOGGER.warning("This packet is not a service 3 packet!")
            return
        hook_obj = get_global_hook_obj()
        if hook_obj is None:
            LOGGER.warning("Hook object not set")
            return
        srv3_packet = cast(Service3Base, packet_if)
        if srv3_packet.custom_hk_handling:
            (
                hk_header,
                hk_content,
                validity_buffer,
                num_vars,
            ) = hook_obj.handle_service_3_housekeeping(
                object_id=bytes(),
                set_id=srv3_packet.set_id,
                hk_data=packet_if.tm_data,
                service3_packet=srv3_packet,
            )
        else:
            (
                hk_header,
                hk_content,
                validity_buffer,
                num_vars,
            ) = hook_obj.handle_service_3_housekeeping(
                object_id=srv3_packet.object_id.as_bytes,
                set_id=srv3_packet.set_id,
                hk_data=packet_if.tm_data[8:],
                service3_packet=srv3_packet,
            )
        if packet_if.subservice == 25 or packet_if.subservice == 26:
            self.handle_hk_print(
                object_id=srv3_packet.object_id.as_int,
                set_id=srv3_packet.set_id,
                hk_header=hk_header,
                hk_content=hk_content,
                validity_buffer=validity_buffer,
                num_vars=num_vars,
            )
        if packet_if.subservice == 10 or packet_if.subservice == 12:
            self.handle_hk_definition_print(
                object_id=srv3_packet.object_id.id,
                set_id=srv3_packet.set_id,
                srv3_packet=srv3_packet,
            )

    def handle_service_5_packet(self, packet_if: PusTmInterface):
        from tmtccmd.config.hook import get_global_hook_obj

        if packet_if.service != 5:
            LOGGER.warning("This packet is not a service 5 packet!")
            return
        hook_obj = get_global_hook_obj()
        if hook_obj is None:
            LOGGER.warning("Hook object not set")
            return
        srv5_packet = cast(Service5TM, packet_if)
        custom_string = hook_obj.handle_service_5_event(
            object_id=srv5_packet.reporter_id_as_bytes,
            event_id=srv5_packet.event_id,
            param_1=srv5_packet.param_1,
            param_2=srv5_packet.param_2,
        )
        self.__print_buffer = custom_string
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def handle_service_8_packet(self, packet_if: PusTmInterface):
        from tmtccmd.config.hook import get_global_hook_obj

        if packet_if.service != PusServices.SERVICE_8_FUNC_CMD:
            LOGGER.warning("This packet is not a service 8 packet!")
            return
        if packet_if.subservice != Srv8Subservices.DATA_REPLY:
            LOGGER.warning(
                f"This packet is not data reply packet with "
                f"subservice {Srv8Subservices.DATA_REPLY}!"
            )
            return
        hook_obj = get_global_hook_obj()
        if hook_obj is None:
            LOGGER.warning("Hook object not set")
            return
        srv8_packet = cast(Service8TM, packet_if)
        if srv8_packet is None:
            LOGGER.warning("Service 8 object is not instance of Service8TM")
            return
        obj_id = srv8_packet.source_object_id_as_bytes
        action_id = srv8_packet.action_id
        header_list, content_list = hook_obj.handle_service_8_telemetry(
            object_id=obj_id, action_id=action_id, custom_data=srv8_packet.custom_data
        )
        obj_id_dict = hook_obj.get_object_ids()
        rep_str = obj_id_dict.get(bytes(obj_id))
        if rep_str is None:
            rep_str = "unknown object"
        print_string = f"Service 8 data reply from {rep_str} with action ID {action_id}"
        self.__print_buffer = print_string
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        self.__print_buffer = header_list
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        self.__print_buffer = content_list
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def handle_hk_print(
        self,
        object_id: int,
        set_id: int,
        hk_header: list,
        hk_content: list,
        validity_buffer: bytearray,
        num_vars: int,
    ):
        """Prints the passed housekeeping packet, if HK printout is enabled and also adds
        it to the internal print buffer.
        :param object_id:   Object ID in integer format
        :param set_id:      Set ID in integer format
        :param hk_header:   Header list
        :param hk_content:  Content list
        :param validity_buffer: Validity buffer bytearray
        :param num_vars: The number of HK variables contained within the set
        :return:
        """
        self.__print_hk(
            content_type=HkContentType.HK,
            object_id=object_id,
            set_id=set_id,
            header=hk_header,
            content=hk_content,
        )
        self.__print_validity_buffer(validity_buffer=validity_buffer, num_vars=num_vars)

    def handle_hk_definition_print(
        self, object_id: int, set_id: int, srv3_packet: Service3Base
    ):
        """
        :param object_id:
        :param set_id:
        :param srv3_packet:
        :return:
        """
        self.__print_buffer = (
            f"HK Definition from Object ID {object_id:#010x} and set ID {set_id}:"
        )
        def_header, def_list = srv3_packet.get_hk_definitions_list()
        self.__print_hk(
            content_type=HkContentType.DEFINITIONS,
            object_id=object_id,
            set_id=set_id,
            header=def_header,
            content=def_list,
        )

    def __handle_short_print(self, packet_if: PusTmInterface):
        self.__print_buffer = (
            "Received TM["
            + str(packet_if.service)
            + ","
            + str(packet_if.subservice)
            + "]"
        )
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def __handle_long_tm_print(
        self, packet_if: PusTmInterface, info_if: PusTmInfoInterface
    ):
        """Main function to print the most important information inside the telemetry
        :param packet_if: Core packet interface
        :param info_if: Information interface
        :return:
        """
        self.__print_buffer = "Received Telemetry: " + info_if.get_print_info()
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        try:
            self.__handle_column_header_print(info_if=info_if)
            self.__handle_tm_content_print(info_if=info_if)
            self.__handle_additional_printout(info_if=info_if)
        except TypeError as error:
            LOGGER.exception(
                f"Type Error when trying to print TM Packet "
                f"[{packet_if.service} , {packet_if.subservice}]"
            )

    def __handle_column_header_print(self, info_if: PusTmInfoInterface):
        header_list = []
        info_if.append_telemetry_column_headers(header_list=header_list)
        self.__print_buffer = str(header_list)
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def __handle_tm_content_print(self, info_if: PusTmInfoInterface):
        """
        :param info_if: Information interface
        :return:
        """
        content_list = []
        info_if.append_telemetry_content(content_list=content_list)
        self.__print_buffer = str(content_list)
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def __handle_additional_printout(self, info_if: PusTmInfoInterface):
        additional_printout = info_if.get_custom_printout()
        if additional_printout != "":
            self.__print_buffer = additional_printout
            LOGGER.info(self.__print_buffer)

    def __print_hk(
        self,
        content_type: HkContentType,
        object_id: int,
        set_id: int,
        header: list,
        content: list,
    ):
        """This function pretty prints HK packets with a given header and content list
        :param content_type: Type of content for HK packet
        :return:
        """
        if content_type == HkContentType.HK:
            print_prefix = "Housekeeping data"
        elif content_type == HkContentType.DEFINITIONS:
            print_prefix = "Housekeeping definitions"
        else:
            print_prefix = "Unknown housekeeping data"
        self.__print_buffer = (
            f"{print_prefix} from Object ID {object_id:#010x} with Set ID {set_id}"
        )
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        if len(content) == 0 or len(header) == 0:
            self.__print_buffer = "Content and header list empty.."
            LOGGER.info(self.__print_buffer)
            self.add_print_buffer_to_file_buffer()
            return
        self.__print_buffer = str(header)
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        self.__print_buffer = str(content)
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def __print_validity_buffer(self, validity_buffer: bytes, num_vars: int):
        """
        :param validity_buffer: Validity buffer in bytes format
        :return:
        """
        if len(validity_buffer) == 0:
            return
        self.__print_buffer = "Valid: "
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()
        self.__handle_validity_buffer_print(
            validity_buffer=validity_buffer, num_vars=num_vars
        )

    def __handle_validity_buffer_print(self, validity_buffer: bytes, num_vars: int):
        """
        :param validity_buffer:
        :param num_vars: Number of parameters
        :return:
        """
        self.__print_buffer = "["
        counter = 0
        for index, byte in enumerate(validity_buffer):
            for bit in range(1, 9):
                if self.bit_extractor(byte, bit) == 1:
                    self.__print_buffer = self.__print_buffer + "Yes"
                else:
                    self.__print_buffer = self.__print_buffer + "No"
                counter += 1
                if counter == num_vars:
                    self.__print_buffer = self.__print_buffer + "]"
                    break
                self.__print_buffer = self.__print_buffer + ", "
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def __handle_wiretapping_packet(
        self, packet_if: PusTmInterface, info_if: PusTmInfoInterface
    ):
        """
        :param packet_if: Core packet interface
        :param info_if: Information interface
        :return:
        """
        if packet_if.service == 2 and (
            packet_if.subservice == 131 or packet_if.subservice == 130
        ):
            self.__print_buffer = (
                f"Wiretapping Packet or Raw Reply from TM [{packet_if.service},"
                f"{packet_if.subservice}]: "
            )
            self.__print_buffer = self.__print_buffer + info_if.get_source_data_string()
            LOGGER.info(self.__print_buffer)
            self.add_print_buffer_to_file_buffer()

    def __handle_event_packet(self, srv_5_tm: Service5TM):
        printout = srv_5_tm.get_custom_printout()
        if printout != "":
            self.__print_buffer += printout
            LOGGER.info(printout)
            self.add_print_buffer_to_file_buffer()

    def print_string(self, string: str, add_cr_to_file_buffer: bool = False):
        """
        Print a string and adds it to the file buffer.
        :param string:
        :param add_cr_to_file_buffer:
        :return:
        """
        self.__print_buffer = string
        LOGGER.info(self.__print_buffer)
        if self.do_print_to_file:
            self.add_print_buffer_to_file_buffer(add_cr_to_file_buffer)

    def add_to_print_string(self, string_to_add: str = ""):
        """Add a specific string to the current print buffer"""
        self.__print_buffer += string_to_add

    def add_print_buffer_to_file_buffer(
        self, additional_newline: bool = False, newline_before: bool = True
    ):
        """
        Add to file buffer. Some options to optimize the output.
        """
        if self.do_print_to_file:
            if additional_newline:
                if newline_before:
                    self.__file_buffer += f"\n{self.__print_buffer}\n"
                else:
                    self.__file_buffer += f"{self.__print_buffer}\n\n"
            else:
                self.__file_buffer += f"{self.__print_buffer}\n"

    def add_print_buffer_to_buffer_list(self):
        """Add the current print buffer to the buffer list"""
        self.file_buffer_list.append(self.__file_buffer)

    def clear_file_buffer(self):
        """Clears the file buffer"""
        self.__file_buffer = ""

    def print_to_file(
        self, log_name: str = "log/tmtc_log.txt", clear_file_buffer: bool = False
    ):
        """
        :param log_name:
        :param clear_file_buffer:
        :return:
        """
        try:
            file = open(log_name, "w")
        except FileNotFoundError:
            LOGGER.info("Log directory does not exists, creating log folder.")
            os.mkdir("log")
            file = open(log_name, "w")
        file.write(self.__file_buffer)
        if clear_file_buffer:
            self.__file_buffer = ""
        LOGGER.info("Log file written to %s", log_name)
        file.close()

    def print_file_buffer_list_to_file(
        self, log_name: str = "log/tmtc_log.txt", clear_list: bool = True
    ):
        """
        Joins the string list and prints it to an output file.
        :param log_name:
        :param clear_list:
        :return:
        """
        try:
            file = open(log_name, "w")
        except FileNotFoundError:
            LOGGER.info("Log directory does not exists, creating log folder.")
            os.mkdir("log")
            file = open(log_name, "w")
        file_buffer = "".join(self.file_buffer_list)
        file.write(file_buffer)
        if clear_list:
            self.file_buffer_list = []
        LOGGER.info("Log file written to %s", log_name)
        file.close()

    @staticmethod
    def bit_extractor(byte: int, position: int):
        """

        :param byte:
        :param position:
        :return:
        """
        shift_number = position + (6 - 2 * (position - 1))
        return (byte >> shift_number) & 1

    def print_telecommand(
        self, tc_packet_obj: PusTelecommand, tc_packet_raw: bytearray = bytes()
    ):
        """
        This function handles the printing of Telecommands. It assumed the packets are sent
        shortly before or after.
        :param tc_packet_obj:
        :param tc_packet_raw:
        :return:
        """
        if self.print_tc:
            if tc_packet_obj is None:
                LOGGER.error("TMTC Printer: Invalid telecommand")
                return
            if self._display_mode == DisplayMode.SHORT:
                self._handle_short_tc_print(tc_packet_obj=tc_packet_obj)
            else:
                self._handle_long_tc_print(tc_packet_obj=tc_packet_obj)

    def _handle_short_tc_print(self, tc_packet_obj: PusTelecommand):
        """
        Brief TC print
        :param tc_packet_obj:
        :return:
        """
        self.__print_buffer = (
            f"Sent TC[{tc_packet_obj.service}, {tc_packet_obj.subservice}] with SSC "
            f"{tc_packet_obj.ssc}"
        )
        LOGGER.info(self.__print_buffer)
        self.add_print_buffer_to_file_buffer()

    def _handle_long_tc_print(self, tc_packet_obj: PusTelecommand):
        """
        Long TC print
        :param tc_packet_obj:
        :return:
        """
        data_print = get_printable_data_string(
            print_format=PrintFormats.HEX, data=tc_packet_obj.app_data
        )
        try:
            self.__print_buffer = (
                f"Telecommand TC[{tc_packet_obj.service}, {tc_packet_obj.subservice}] "
                f"with SSC {tc_packet_obj.ssc} sent with data "
                f"{data_print}"
            )
            LOGGER.info(self.__print_buffer)
            self.add_print_buffer_to_file_buffer()
        except TypeError as error:
            LOGGER.error("TMTC Printer: Type Error! Traceback: %s", error)

    @staticmethod
    def print_data(data: bytes):
        """
        :param data: Data to print
        :return: None
        """
        string = get_printable_data_string(print_format=PrintFormats.HEX, data=data)
        LOGGER.info(string)



#!/usr/bin/env python3
"""
@file           tmtc_frontend.py
@date           01.11.2019
@brief          This is part of the TMTC client developed by the SOURCE project by KSat
@description    GUI is still work-in-progress
@manual
@author         R. Mueller, P. Scheurenbrand, D. Nguyen
"""
import enum
import os
import sys
import time
import webbrowser
from multiprocessing import Process
from typing import Union

from PyQt5.QtWidgets import (
    QMainWindow,
    QGridLayout,
    QTableWidget,
    QWidget,
    QLabel,
    QCheckBox,
    QDoubleSpinBox,
    QFrame,
    QComboBox,
    QPushButton,
    QTableWidgetItem,
    QMenu,
    QAction,
    QMenuBar,
)
from PyQt5.QtGui import QPixmap, QIcon
from PyQt5.QtCore import Qt, pyqtSignal, QObject, QThread, QRunnable

from tmtccmd.core.frontend_base import FrontendBase
from tmtccmd.core.backend import TmTcHandler
from tmtccmd.config.hook import TmTcHookBase
from tmtccmd.config.definitions import CoreGlobalIds, CoreModeList, CoreComInterfaces
from tmtccmd.config.hook import get_global_hook_obj
from tmtccmd.utility.logger import get_console_logger
from tmtccmd.core.globals_manager import get_global, update_global
from tmtccmd.com_if.tcpip_utilities import TcpIpConfigIds
import tmtccmd.config as config_module


LOGGER = get_console_logger()


CONNECT_BTTN_STYLE = (
    "background-color: #1fc600;"
    "border-style: inset;"
    "font: bold;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


DISCONNECT_BTTN_STYLE = (
    "background-color: orange;"
    "border-style: inset;"
    "font: bold;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


COMMAND_BUTTON_STYLE = (
    "background-color: #cdeefd;"
    "border-style: inset;"
    "font: bold;"
    "padding: 6px;"
    "border-width: 2px;"
    "border-radius: 6px;"
)


class WorkerOperationsCodes(enum.IntEnum):
    DISCONNECT = 0
    SEQUENTIAL_COMMANDING = 1


class WorkerThread(QObject):
    finished = pyqtSignal()

    def __init__(self, op_code: WorkerOperationsCodes, tmtc_handler: TmTcHandler):
        super(QObject, self).__init__()
        self.op_code = op_code
        self.tmtc_handler = tmtc_handler

    def run_worker(self):
        if self.op_code == WorkerOperationsCodes.DISCONNECT:
            self.tmtc_handler.close_listener()
            while True:
                if not self.tmtc_handler.is_com_if_active():
                    break
                else:
                    time.sleep(0.4)
            self.finished.emit()
        elif self.op_code == WorkerOperationsCodes.SEQUENTIAL_COMMANDING:
            self.tmtc_handler.set_mode(CoreModeList.SEQUENTIAL_CMD_MODE)
            # It is expected that the TMTC handler is in the according state to perform the operation
            self.tmtc_handler.perform_operation()
            self.finished.emit()
        else:
            LOGGER.warning("Unknown worker operation code!")
            self.finished.emit()


class RunnableThread(QRunnable):
    """
    Runnable thread which can be used with QThreadPool. Not used for now, might be needed in the future.
    """

    def run(self):
        pass


class TmTcFrontend(QMainWindow, FrontendBase):
    def __init__(
        self, hook_obj: TmTcHookBase, tmtc_backend: TmTcHandler, app_name: str
    ):
        super(TmTcFrontend, self).__init__()
        super(QMainWindow, self).__init__()
        self._tmtc_handler = tmtc_backend
        self._app_name = app_name
        self._hook_obj = hook_obj

        self._tmtc_handler.initialize()
        self.service_op_code_dict = dict()
        self._service_list = []
        self._op_code_list = []
        self._com_if_list = []
        self._last_com_if = CoreComInterfaces.UNSPECIFIED.value
        self._current_com_if = CoreComInterfaces.UNSPECIFIED.value
        self._current_service = ""
        self._current_op_code = ""
        self._current_com_if_key = "unspec"
        self.__connected = False

        self.__worker = None
        self.__thread = None
        self.__debug_mode = False

        self.__combo_box_op_codes: Union[None, QComboBox] = None
        module_path = os.path.abspath(config_module.__file__).replace("__init__.py", "")
        self.logo_path = f"{module_path}/logo.png"

    def prepare_start(self, args: any) -> Process:
        return Process(target=self.start)

    def start(self, qt_app: any):
        self.__start_ui()
        sys.exit(qt_app.exec())

    def set_gui_logo(self, logo_total_path: str):
        if os.path.isfile(logo_total_path):
            self.logo_path = logo_total_path
        else:
            LOGGER.warning("Could not set logo, path invalid!")

    def __start_ui(self):
        self.__create_menu_bar()
        win = QWidget(self)
        self.setCentralWidget(win)

        grid = QGridLayout()
        win.setLayout(grid)
        row = 0
        self.setWindowTitle(self._app_name)
        self.setWindowIcon(QIcon(self.logo_path))

        add_pixmap = False

        if add_pixmap:
            row = self.__set_up_pixmap(grid=grid, row=row)

        row = self.__set_up_config_section(grid=grid, row=row)
        row = self.__add_vertical_separator(grid=grid, row=row)

        # com if configuration
        row = self.__set_up_com_if_section(grid=grid, row=row)
        row = self.__add_vertical_separator(grid=grid, row=row)

        row = self.__set_up_service_op_code_section(grid=grid, row=row)

        self.__command_button = QPushButton()
        self.__command_button.setText("Send Command")
        self.__command_button.setStyleSheet(COMMAND_BUTTON_STYLE)
        self.__command_button.clicked.connect(self.__start_seq_cmd_op)
        self.__command_button.setEnabled(False)
        grid.addWidget(self.__command_button, row, 0, 1, 2)
        row += 1
        self.show()

    def __start_seq_cmd_op(self):
        if self.__debug_mode:
            LOGGER.info("Start Service Test Button pressed.")
        if not self.__get_send_button():
            return
        self.__set_send_button(False)
        self._tmtc_handler.set_service(self._current_service)
        self._tmtc_handler.set_opcode(self._current_op_code)
        self.__start_qthread_task(
            op_code=WorkerOperationsCodes.SEQUENTIAL_COMMANDING,
            finish_callback=self.__finish_seq_cmd_op,
        )

    def __finish_seq_cmd_op(self):
        self.__set_send_button(True)

    def __connect_button_action(self):
        if not self.__connected:
            LOGGER.info("Starting TM listener..")
            # Build and assign new communication interface
            if self._current_com_if != self._last_com_if:
                hook_obj = get_global_hook_obj()
                new_com_if = hook_obj.assign_communication_interface(
                    com_if_key=self._current_com_if,
                    tmtc_printer=self._tmtc_handler.get_printer(),
                )
                self._tmtc_handler.set_com_if(new_com_if)
            self._tmtc_handler.start_listener(False)
            self.__connect_button.setStyleSheet(DISCONNECT_BTTN_STYLE)
            self.__command_button.setEnabled(True)
            self.__connect_button.setText("Disconnect")
            self.__connected = True
        else:
            LOGGER.info("Closing TM listener..")
            self.__command_button.setEnabled(False)
            self.__connect_button.setEnabled(False)
            self.__start_qthread_task(
                op_code=WorkerOperationsCodes.DISCONNECT,
                finish_callback=self.__finish_disconnect_button_op,
            )

    def __finish_disconnect_button_op(self):
        self.__connect_button.setEnabled(True)
        # self.__disconnect_button.setEnabled(False)
        self.__connect_button.setStyleSheet(CONNECT_BTTN_STYLE)
        self.__connect_button.setText("Connect")
        LOGGER.info("Disconnect successfull")
        self.__connected = False

    def __create_menu_bar(self):
        menu_bar = self.menuBar()
        # Creating menus using a QMenu object
        file_menu = QMenu("&File", self)
        menu_bar.addMenu(file_menu)
        # Creating menus using a title
        help_menu = menu_bar.addMenu("&Help")

        help_action = QAction("Help", self)
        help_action.triggered.connect(self.__help_url)
        help_menu.addAction(help_action)

    @staticmethod
    def __help_url():
        webbrowser.open("https://tmtccmd.readthedocs.io/en/latest/")

    def __set_up_config_section(self, grid: QGridLayout, row: int) -> int:
        grid.addWidget(QLabel("Configuration:"), row, 0, 1, 2)
        row += 1
        checkbox_console = QCheckBox("Print output to console")
        checkbox_console.stateChanged.connect(self.__checkbox_console_update)

        checkbox_log = QCheckBox("Print output to log file")
        checkbox_log.stateChanged.connect(self.__checkbox_log_update)

        checkbox_raw_tm = QCheckBox("Print all raw TM data directly")
        checkbox_raw_tm.stateChanged.connect(self.__checkbox_print_raw_data_update)

        checkbox_hk = QCheckBox("Print Housekeeping Data")
        # checkbox_hk.setChecked(tmtcc_config.G_PRINT_HK_DATA)
        checkbox_hk.stateChanged.connect(checkbox_print_hk_data)

        checkbox_short = QCheckBox("Short Display Mode")
        # checkbox_short.setChecked(tmtcc_config.G_DISPLAY_MODE == "short")
        checkbox_short.stateChanged.connect(checkbox_short_display_mode)

        grid.addWidget(checkbox_log, row, 0, 1, 1)
        grid.addWidget(checkbox_console, row, 1, 1, 1)
        row += 1
        grid.addWidget(checkbox_raw_tm, row, 0, 1, 1)
        grid.addWidget(checkbox_hk, row, 1, 1, 1)
        row += 1
        grid.addWidget(checkbox_short, row, 0, 1, 1)
        row += 1

        grid.addWidget(QLabel("TM Timeout:"), row, 0, 1, 1)
        grid.addWidget(QLabel("TM Timeout Factor:"), row, 1, 1, 1)
        row += 1

        spin_timeout = QDoubleSpinBox()
        spin_timeout.setValue(4)
        # TODO: set sensible min/max values
        spin_timeout.setSingleStep(0.1)
        spin_timeout.setMinimum(0.25)
        spin_timeout.setMaximum(60)
        # https://youtrack.jetbrains.com/issue/PY-22908
        # Ignore those warnings for now.
        spin_timeout.valueChanged.connect(number_timeout)
        grid.addWidget(spin_timeout, row, 0, 1, 1)

        spin_timeout_factor = QDoubleSpinBox()
        # spin_timeout_factor.setValue(tmtcc_config.G_TC_SEND_TIMEOUT_FACTOR)
        # TODO: set sensible min/max values
        spin_timeout_factor.setSingleStep(0.1)
        spin_timeout_factor.setMinimum(0.25)
        spin_timeout_factor.setMaximum(10)
        spin_timeout_factor.valueChanged.connect(number_timeout_factor)
        grid.addWidget(spin_timeout_factor, row, 1, 1, 1)
        row += 1
        return row

    def __set_up_com_if_section(self, grid: QGridLayout, row: int) -> int:
        grid.addWidget(QLabel("Communication Interface:"), row, 0, 1, 1)
        com_if_combo_box = QComboBox()
        all_com_ifs = get_global(CoreGlobalIds.COM_IF_DICT)
        index = 0
        # add all possible ComIFs to the comboBox
        for com_if_key, com_if_value in all_com_ifs.items():
            com_if_combo_box.addItem(com_if_value)
            self._com_if_list.append((com_if_key, com_if_value))
            if self._tmtc_handler.get_com_if_id() == com_if_key:
                com_if_combo_box.setCurrentIndex(index)
            index += 1
        com_if_combo_box.currentIndexChanged.connect(self.__com_if_index_changed)
        grid.addWidget(com_if_combo_box, row, 1, 1, 1)
        row += 1

        self.com_if_cfg_button = QPushButton()
        self.com_if_cfg_button.setText("Configure")
        grid.addWidget(self.com_if_cfg_button, row, 0, 1, 2)
        row += 1

        self.__connect_button = QPushButton()
        self.__connect_button.setText("Connect")
        self.__connect_button.setStyleSheet(CONNECT_BTTN_STYLE)
        self.__connect_button.clicked.connect(self.__connect_button_action)

        grid.addWidget(self.__connect_button, row, 0, 1, 2)
        row += 1
        return row

    def __set_up_service_op_code_section(self, grid: QGridLayout, row: int):
        grid.addWidget(QLabel("Service: "), row, 0, 1, 2)
        grid.addWidget(QLabel("Operation Code: "), row, 1, 1, 2)
        row += 1

        combo_box_services = QComboBox()
        default_service = get_global(CoreGlobalIds.CURRENT_SERVICE)
        self.service_op_code_dict = self._hook_obj.get_service_op_code_dictionary()
        if self.service_op_code_dict is None:
            LOGGER.warning("Invalid service to operation code dictionary")
            LOGGER.warning("Setting default dictionary")
            from tmtccmd.config.globals import get_default_service_op_code_dict

            self.service_op_code_dict = get_default_service_op_code_dict()
        index = 0
        default_index = 0
        for service_key, service_value in self.service_op_code_dict.items():
            combo_box_services.addItem(service_value[0])
            if service_key == default_service:
                default_index = index
            self._service_list.append(service_key)
            index += 1
        combo_box_services.setCurrentIndex(default_index)
        self._current_service = self._service_list[default_index]

        combo_box_services.currentIndexChanged.connect(self.__service_index_changed)
        grid.addWidget(combo_box_services, row, 0, 1, 1)

        self.__combo_box_op_codes = QComboBox()
        self._current_service = self._service_list[default_index]
        self.__update_op_code_combo_box()
        self.__combo_box_op_codes.currentIndexChanged.connect(
            self.__op_code_index_changed
        )
        # TODO: Combo box also needs to be updated if another service is selected
        grid.addWidget(self.__combo_box_op_codes, row, 1, 1, 1)
        row += 1
        return row

    def __set_up_pixmap(self, grid: QGridLayout, row: int) -> int:
        label = QLabel(self)
        label.setGeometry(720, 10, 100, 100)
        label.adjustSize()

        pixmap = QPixmap(self.logo_path)
        pixmap_width = pixmap.width()
        pixmap_height = pixmap.height()
        row += 1

        pixmap_scaled = pixmap.scaled(
            pixmap_width * 0.3, pixmap_height * 0.3, Qt.KeepAspectRatio
        )
        label.setPixmap(pixmap_scaled)
        label.setScaledContents(True)

        grid.addWidget(label, row, 0, 1, 2)
        row += 1
        return row

    def __start_qthread_task(self, op_code: WorkerOperationsCodes, finish_callback):
        self.__thread = QThread()
        self.__worker = WorkerThread(op_code=op_code, tmtc_handler=self._tmtc_handler)
        self.__worker.moveToThread(self.__thread)

        self.__thread.started.connect(self.__worker.run_worker)
        self.__worker.finished.connect(self.__thread.quit)
        self.__worker.finished.connect(self.__worker.deleteLater)
        self.__thread.finished.connect(self.__thread.deleteLater)

        self.__thread.finished.connect(finish_callback)
        self.__thread.start()

    @staticmethod
    def __add_vertical_separator(grid: QGridLayout, row: int):
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        grid.addWidget(separator, row, 0, 1, 2)
        row += 1
        return row

    def __service_index_changed(self, index: int):
        self._current_service = self._service_list[index]
        self.__update_op_code_combo_box()
        if self.__debug_mode:
            LOGGER.info("Service changed")

    def __op_code_index_changed(self, index: int):
        self._current_op_code = self._op_code_list[index]
        if self.__debug_mode:
            LOGGER.info("Op Code changed")

    def __update_op_code_combo_box(self):
        self.__combo_box_op_codes.clear()
        self._op_code_list = []
        op_code_dict = self.service_op_code_dict[self._current_service][1]
        if op_code_dict is not None:
            for op_code_key, op_code_value in op_code_dict.items():
                try:
                    self._op_code_list.append(op_code_key)
                    self.__combo_box_op_codes.addItem(op_code_value[0])
                except TypeError:
                    LOGGER.warning(f"Invalid op code entry {op_code_value}, skipping..")
            self._current_op_code = self._op_code_list[0]

    def __checkbox_log_update(self, state: int):
        update_global(CoreGlobalIds.PRINT_TO_FILE, state)
        if self.__debug_mode:
            LOGGER.info(["Enabled", "Disabled"][state == 0] + " print to log.")

    def __checkbox_console_update(self, state: bool):
        update_global(CoreGlobalIds.PRINT_TM, state)
        if self.__debug_mode:
            LOGGER.info(["enabled", "disabled"][state == 0] + " console print")

    def __checkbox_print_raw_data_update(self, state: int):
        update_global(CoreGlobalIds.PRINT_RAW_TM, state)
        if self.__debug_mode:
            LOGGER.info(["enabled", "disabled"][state == 0] + " printing of raw data")

    def __set_send_button(self, state: bool):
        self.__command_button.setEnabled(state)

    def __get_send_button(self):
        return self.__command_button.isEnabled()

    def __com_if_index_changed(self, index: int):
        self._current_com_if = self._com_if_list[index][0]
        if self.__debug_mode:
            LOGGER.info(f"Communication IF updated: {self._com_if_list[index][1]}")


class SingleCommandTable(QTableWidget):
    def __init__(self):
        super().__init__()
        self.setRowCount(1)
        self.setColumnCount(5)
        self.setHorizontalHeaderItem(0, QTableWidgetItem("Service"))
        self.setHorizontalHeaderItem(1, QTableWidgetItem("Subservice"))
        self.setHorizontalHeaderItem(2, QTableWidgetItem("SSC"))
        self.setHorizontalHeaderItem(3, QTableWidgetItem("Data"))
        self.setHorizontalHeaderItem(4, QTableWidgetItem("CRC"))
        self.setItem(0, 0, QTableWidgetItem("17"))
        self.setItem(0, 1, QTableWidgetItem("1"))
        self.setItem(0, 2, QTableWidgetItem("20"))


def checkbox_print_hk_data(state: int):
    update_global(CoreGlobalIds.PRINT_HK, state)
    LOGGER.info(["enabled", "disabled"][state == 0] + " printing of hk data")


def checkbox_short_display_mode(state: int):
    update_global(CoreGlobalIds.DISPLAY_MODE, state)
    LOGGER.info(["enabled", "disabled"][state == 0] + " short display mode")


def number_timeout(value: float):
    update_global(CoreGlobalIds.TM_TIMEOUT, value)
    LOGGER.info("PUS TM timeout changed to: " + str(value))


def number_timeout_factor(value: float):
    update_global(CoreGlobalIds.TC_SEND_TIMEOUT_FACTOR, value)
    LOGGER.info("PUS TM timeout factor changed to: " + str(value))


def ip_change_client(value):
    ethernet_config = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    ethernet_config[TcpIpConfigIds.RECV_ADDRESS] = value
    update_global(CoreGlobalIds.ETHERNET_CONFIG, ethernet_config)
    LOGGER.info("Client IP changed: " + value)


def ip_change_board(value):
    ethernet_config = get_global(CoreGlobalIds.ETHERNET_CONFIG)
    ethernet_config[TcpIpConfigIds.SEND_ADDRESS] = value
    update_global(CoreGlobalIds.ETHERNET_CONFIG, ethernet_config)
    LOGGER.info("Board IP changed: " + value)



from typing import Callable, Dict, Optional, Tuple, List

from tmtccmd.tm.handler import TmHandler
from tmtccmd.tm.definitions import TelemetryQueueT, TmTypes
from tmtccmd.sendreceive.tm_listener import QueueListT
from tmtccmd.utility.tmtc_printer import TmTcPrinter
from tmtccmd.utility.logger import get_console_logger

CcsdsCallbackT = Callable[[int, bytearray, TmTcPrinter], None]

LOGGER = get_console_logger()
HandlerDictT = Dict[int, Tuple[CcsdsCallbackT, int]]


class CcsdsTmHandler(TmHandler):
    """Generic CCSDS handler class. The user can create an instance of this class to handle
    CCSDS packets with different APIDs"""

    def __init__(self, tmtc_printer: Optional[TmTcPrinter] = None):
        super().__init__(tm_type=TmTypes.CCSDS_SPACE_PACKETS)
        self._handler_dict: HandlerDictT = dict()
        self._tmtc_printer = tmtc_printer

    def initialize(self, tmtc_printer: TmTcPrinter):
        self._tmtc_printer = tmtc_printer

    def add_tm_handler(
        self, apid: int, pus_tm_handler: CcsdsCallbackT, max_queue_len: int
    ):
        """Add a TM handler for a certain APID. The handler is a callback function which
        will be called if telemetry with that APID arrives
        :param apid:            CCSDS Application Process ID
        :param pus_tm_handler:  Callback function
        :param max_queue_len:
        :return:
        """
        self._handler_dict[apid] = pus_tm_handler, max_queue_len

    def get_apid_queue_len_list(self) -> List[Tuple[int, int]]:
        list = []
        for apid, handler_tuple in self._handler_dict.items():
            list.append((apid, handler_tuple[1]))
        return list

    def handle_packet_queues(self, packet_queue_list: QueueListT):
        for queue_tuple in packet_queue_list:
            apid = queue_tuple[0]
            handler_tuple = self._handler_dict.get(apid)
            if handler_tuple is not None:
                ccsds_cb = handler_tuple[0]
                self.handle_ccsds_packet_queue(
                    apid=apid, packet_queue=queue_tuple[1], ccsds_cb=ccsds_cb
                )

    def handle_ccsds_packet_queue(
        self,
        apid: int,
        packet_queue: TelemetryQueueT,
        ccsds_cb: Optional[CcsdsCallbackT] = None,
    ):
        for tm_packet in packet_queue:
            if ccsds_cb is None:
                ccsds_cb = self._handler_dict[apid][0]
            ccsds_cb(apid, tm_packet, self._tmtc_printer)
